// $Id: Game2EventHandler.C,v 1.12 2006/12/31 04:09:34 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

/* 
   Copyright 2007 Man Hon Chan, Zhangbo Liu, Simon Suyadi
   This code is Licensed under GPL.
 */

// A sample client for ORTS that shows how to access client data

// todo/questions:
//
// game scripts:
// - what to do with workers during build time
// - similarly, how to define mining?
// - how is attack range defined (distance between centers/hulls)?
//
// client:
// - how to access global player data?
// - how to compute the set of visible tiles?
// - how to compute tiles visible from a unit?
// - how to decide what can be attacked by a specific unit?

// FORS/FORALL/ERR macros are defined in orts3/libs/kernel/src/Global.H

#include "DebugHeader.H"
#include "Game2EventHandler.H"
#include "GameStateModule.H"
#include "GfxModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"
#include "GUI.H"
#include "DrawOnTerrain2D.H"
#include <algorithm>
#include <cstdio>
#include <iostream>
#include "FFEngine.H"

DEF_FERR("GameEventHandler");

// tile types from core.bp
// fixme: game script reflection would help here
// in the competition we won't use water tiles
// TILE_CLIFF: non-traversable by ground units

namespace {

  static const int NUMBER_OF_SQUADS = 10;
  static const int NUMBER_OF_TANKS = 50;
  static const int NUMBER_OF_BASES = 5;
  const int FF_INTERVAL = 2;
  const int GENERAL_AI_UPDATE_INTERVAL = 8;

  // For GUI
  const Vec3<real4> WHITE(1.0, 1.0, 1.0);
  const Vec3<real4> RED(1.0, 0.0, 0.0);
  const Vec3<real4> GREEN(0.0, 1.0, 0.0);
  const Vec3<real4> BLUE(0.0, 0.0, 1.0);
  const Vec3<real4> YELLOW(1.0, 1.0, 0.0);
  const Vec3<real4> CYAN(0.0, 1.0, 1.0);
  const Vec3<real4> MAGENTA(1.0, 0.0, 1.0);
  const Vec3<real4> GREY(0.5, 0.5, 0.5);
  const Vec3<real4> BLACK(0.0, 0.0, 0.0);
}

#define FOR_RANGE(it, begin, end) for ( typeof(begin) it = (begin),\
    _end_of_loop_ ## it = (end); it != _end_of_loop_ ## it ; ++it)

bool Game2EventHandler::handle_event(const Event& e)
{

  // Post initialize
  if (!postInitialized) {
    PostInitialize();
    LOG( ferr << "Post Init done" << endl );
  }

  if (e.get_who() == GameStateModule::FROM) {

    //cout << "Event: " << e.who() << " " << e.what() << endl;

    if (e.get_what() == GameStateModule::STOP_MSG) {
      state.quit = true; return true;
    } else if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
      state.quit = state.error = true; return true;
    } else if (e.get_what() == GameStateModule::VIEW_MSG) {

      LOG(ferr << "Before Update" << endl);
      GlobalDataBase::Update(state.gsm->get_changes());
      LOG(ferr << "Before Clustering::Update()" << endl);
      Clustering::Update();

      // Check who to fire

      sint4 game_tick = GlobalDataBase::GetGameTick();

      if (game_tick - prevGeneralAITick >= GENERAL_AI_UPDATE_INTERVAL) {
        prevGeneralAITick = game_tick;
        LOG(ferr << "Before GeneralAI::Update()" << endl);
        generalAI->Update();
      }

      LOG(ferr << "Before Compute Action" << endl);
      ComputeActions();
      if (game_tick - prevFFTick >= FF_INTERVAL) {
        // Move
        prevFFTick = game_tick;
        LOG(ferr << "Before FFEngine::Run()" << endl);
        FFEngine::Run();
      }

      LOG(ferr << "Before Act" << endl);
      // Set all actions
      PlayerTank** tank = GlobalDataBase::GetPlayerTanks();
      PlayerTank** tank_end = GlobalDataBase::GetPlayerTanksEnd();
      for(; tank != tank_end; ++tank) (*tank)->Act();
      LOG(ferr << "Before send action" << endl);
      // All actions set, now send them
      state.gsm->send_actions();

      // REMARK: Here belows are debugging information

      Game &game = state.gsm->get_game();	
      // gfx and frame statistics

      if (state.gfxm) state.gfxm->process_changes();

      sint4 vf = game.get_view_frame();
      sint4 af = game.get_action_frame();
      sint4 sa = game.get_skipped_actions();
      sint4 ma = game.get_merged_actions();

      if ((vf % 100) == 0) {
        GlobalDataBase::Display();
      }

      if ((vf % 10) == 0) {
        printf("%05d\n", vf);
        //cout << "[frame " << vf << "]" << endl;
      }

      if (vf != af || sa || ma > 1) {
        cout << "frame " << vf;
        if (af < 0) cout << " [no action]";
        else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
        if (sa) cout << " [skipped " << sa << "]";
        if (ma > 1) cout << " [merged " << ma << "]";
        cout << endl;
      }

      if (state.gfxm) {

        // don't draw if we are far behind
        if (abs(vf-af) < 10) {
          state.gfxm->draw();
          state.just_drew = true;
        }
      }

      if (state.gui) {
        state.gui->event();
        if (state.gui->quit) exit(0);
      }

      if (state.gui) {
        // as a hack for flickering, gui->display() is called in dot.start()
        // so call everything that draws to the debug window between start() and end()
        DrawOnTerrain2D dot(state.gui);
        dot.start();

        // Draw visibility graph
        /*
        for (const PlayerTank* const* i = GlobalDataBase::GetPlayerTanks(), * const* ei = GlobalDataBase::GetPlayerTanksEnd(); i != ei; ++i)
          // Go through all enemy objects
          for (const EnemyObj* const* j = GlobalDataBase::GetEnemyUnits(),
              * const* ej = GlobalDataBase::GetEnemyUnitsEnd(); j != ej; ++j)
            if ( GlobalDataBase::CanSee(*i, *j) ) {
              const WorldCoor &player_coor = (*i)->GetWorldCoor();
              const WorldCoor &enemy_coor = (*j)->GetWorldCoor();
              dot.draw_line(player_coor.x, player_coor.y,
                  enemy_coor.x, enemy_coor.y, GREY);
            }
            */

        Cluster* const* cluster_begin = Clustering::GetClusters();
        Cluster* const* cluster_end = Clustering::GetClustersEnd();
        for (Cluster* const* i = cluster_begin; i != cluster_end; ++i) {
          Cluster* const cluster = *i;
          // Get the hulls...
          WorldCoor prev_coor = cluster->GetHullEnd()[-1];
          for (const WorldCoor* j = cluster->GetHull(), *ej = cluster->GetHullEnd();
              j != ej; ++j) {
            const WorldCoor &now = *j;
            dot.draw_line(prev_coor.x, prev_coor.y, now.x, now.y, WHITE);
            prev_coor = now;
          }
          WorldCoor center = cluster->GetCenterOfMass();
          dot.draw_circle(center.x, center.y, 5, MAGENTA);
          dot.draw_circle(center.x, center.y, 10, MAGENTA);
        }

        for (int r = 0; r < 64; ++r)
          for (int c = 0; c < 64; ++c) {
            MapCoor coor(r,c);
            if (MapBox::Walkable(coor)) continue;
            int left = MapBox::GetLeft(coor);
            int right = MapBox::GetRight(coor);
            int top = MapBox::GetTop(coor);
            int bottom = MapBox::GetBottom(coor);
            dot.draw_line(left, top, right, bottom, MAGENTA, 3);
            dot.draw_line(left, bottom, right, top, MAGENTA, 3);
          }

        /*
           for (const EnemyTank* const* i = GlobalDataBase::GetEnemyTanks(),
         * const* ei = GlobalDataBase::GetEnemyTanksEnd(); i != ei; ++i)
        // Go through all enemy objects
        for (const EnemyObj* const* j = GlobalDataBase::GetEnemyUnits(),
         * const* ej = GlobalDataBase::GetEnemyUnitsEnd(); j != ej; ++j)
         if ( GlobalDataBase::CanSee(*i, *j) ) {
         const WorldCoor &tank_coor = (*i)->GetWorldCoor();
         const WorldCoor &obj_coor = (*j)->GetWorldCoor();
         dot.draw_line(tank_coor.x, tank_coor.y,
         obj_coor.x, obj_coor.y, MAGENTA);
         }
         */

        // Draw all directions
        for (const PlayerTank* const* i = GlobalDataBase::GetPlayerTanks(),
            * const* ei = GlobalDataBase::GetPlayerTanksEnd(); i != ei; ++i) {
          const PlayerTank* tank = *i;

          {
            const WorldCoor& now = tank->GetWorldCoor();
            const WorldCoor& way_point = tank->GetWayPoint();
            const int dx = way_point.x - now.x;
            const int dy = way_point.y - now.y;
            const WorldCoor head(now.x + 5 * dx, now.y + 5 * dy);
            dot.draw_line(now.x, now.y, head.x, head.y, YELLOW);
          }

          {
            const WorldCoor& now = tank->GetWorldCoor();
            const WorldCoor& ctr = tank->GetSquad()->GetCenterOfMass();
            dot.draw_line(now.x, now.y, ctr.x, ctr.y, GREY);
          }

          {
            const WorldCoor& now = tank->GetWorldCoor();
            const WorldCoor ctr(tank->GetSquad()->GetDest());
            dot.draw_line(now.x, now.y, ctr.x, ctr.y, BLUE);
          }
        }

        dot.end();
      }

      return true;
    }
  }

  return false;
}

namespace {
  const sint4 SQR_RANGE = 110 * 110;
}

void Game2EventHandler::ComputeActions()
{
  generalAI->Act();
  return;
}

// Handle everything that depends on the game
void Game2EventHandler::PostInitialize() {
  Game &game = state.gsm->get_game();

  GlobalDataBase::InitGame(&game);
  FFEngine::Init();
  LOG(cerr << "Start init GeneralAI..." << endl);
  generalAI = new GeneralAI();
  generalAI->Init();
  LOG(cerr << "Finish init GeneralAI..." << endl);

  postInitialized = true;
}
