#include "DebugHeader.H"
#include "GeneralAI.H"
#include <algorithm>
#include <cmath>

/* 
   Copyright 2007 Man Hon Chan, Zhangbo Liu, Simon Suyadi
   This code is Licensed under GPL.
 */

DEF_FERR("GeneralAI");

using std::max;
using std::min;
using std::size_t;
using std::abs;

namespace {
  const sint4 INITIAL_SQUAD_NUMBER = 5;
  const sint4 SQUAD_MERGE_DIST = 64 * 64 * 9 / 2 / 2;
  const sint4 SQUAD_TANK_MERGE_DIST = 16 * 16;

  const MapCoor NO_TABOO(-1, -1);
  int BIT_COUNT[1 << MAX_SQUAD_NUMBER];

  int StaticInit() {
    BIT_COUNT[0] = 0;
    int n = 1 << MAX_SQUAD_NUMBER;
    for (int i = 1; i < n; ++i)
      BIT_COUNT[i] = (i & 1) + BIT_COUNT[i >> 1];
    return 0;
  }

  const int dummy = StaticInit();

  // squad target decision
  const float SQUAD_AVG_FIREPOWER = 30.0f / 20;
  const float SQUAD_AVG_SPEED = 2.7f;
  const float SQUAD_RESCUE_MULT = 1.2f;
  const int RESCUE_TIME_OFFSET = 16;
  const int RESCUE_DIST = 16;
  const int TARGET_FAR_SCAN_SQR_RANGE = (112 * 3 / 2) * (112 * 3 / 2);
  // formula for finding then number of our tanks to send
  inline int RescueSquadCount(const int& enemy) {
    int n = (int)ceil(enemy * SQUAD_RESCUE_MULT + 1);
    return max(n, 4);
  }
}

#define FOR_RANGE(it, begin, end) for ( typeof(begin) it = (begin),\
    _end_of_loop_ ## it = (end); it != _end_of_loop_ ## it ; ++it)

#define CLEAR_DEAD(v) for (size_t i = 0; i < v.size(); ) if (!v[i]->IsAlive()) { v[i] = v.back(); v.pop_back(); } else ++i;

GeneralAI::GeneralAI() : aiMode(AI_DEFAULT) {
  memset(engagingMap, 0, MAX_SQUAD_NUMBER * MAX_SQUAD_NUMBER); 
  tabooCount = 0;

  // TEMPORARY
  //aiMode = AI_DEFENSIVE;
}

GeneralAI::~GeneralAI() {
  //TODO: delete all squads here
  for (size_t i = 0; i < squadAI.size(); i++) { 
    delete squadAI[i];
  }
}

int GeneralAI::CalcRescueTime(PlayerCC* c, SquadAI* squad, const int& req) {
  LOG(ferr << "Within CalcRescueTime" << endl);
  //const int& enemyhp = c->GetHarassTankTotalHP();
  const int& hp = c->GetHP();
  vector<PlayerTank*> &tanks = squad->GetTanks();
  const int& enemy_count =c->GetHarassTankCount(); 
  const int& N = tanks.size();

  // time for the CC to die
  float timez = (float)hp / (enemy_count * SQUAD_AVG_FIREPOWER);

  // will never kill the enemy
  if(N < req) return -1;

  // can use centre of mass to estimate
  vector<real4> dists(tanks.size());
  for(size_t i = 0; i < tanks.size(); i++) {
    dists[i] = tanks[i]->GetDist(c);
  }
  sort(dists.begin(), dists.end());
  float avg_dist = 0; //average distance
  for(int i = 0; i < enemy_count; i++) {
    avg_dist += dists[i];
  }
  avg_dist /= enemy_count;

  // time to reach destination
  timez -= avg_dist / SQUAD_AVG_SPEED;
  //timez -= (enemyhp / (SQUAD_AVG_FIREPOWER * N));
  LOG(ferr << "End of CaleRescueTimer: " << timez << endl);
  return (int)timez;
}

// what happens if we move the whole bunch of squad to this point?
int GeneralAI::CalcPoint(SquadAI* squad, const WorldCoor& dest) {
  const int CC_PT = 10000; // a very high point for this
  const int ENEMY_PT = 10; // points for engaing enemy

  int pt = 0;
  // sum up the 'force' from the CCs
  FOR_RANGE(cc, GlobalDataBase::GetPlayerCCs(), GlobalDataBase::GetPlayerCCsEnd()) {
    if ((*cc)->IsDead()) continue;
    float dist = Dist(dest, (*cc)->GetWorldCoor());
    float ccpt = CC_PT / (dist * dist); //squared distance
  }
  LOG( ferr << "Points from CCs=" << pt << endl );
}

void GeneralAI::Init() {
  //1. Initially assign tanks to squads
  LOG(ferr << "Before everything in Init()" << endl);
  for (int i = 0; i < INITIAL_SQUAD_NUMBER; i++) {
    SquadAI* squad_ai = new SquadAI(i); // i is the ID
    squadAI.push_back(squad_ai);
  }
  //Decide which tank goes to which squad
  LOG(ferr << "Decide which tank goes to which squad" << endl);
  PlayerTank** player_tank = GlobalDataBase::GetPlayerTanks();
  PlayerTank** player_tank_end = GlobalDataBase::GetPlayerTanksEnd();
  for (PlayerTank** i = player_tank; i != player_tank_end; ++i) {
    int best_id = 0;
    for (int j = 1; j < GlobalDataBase::GetPlayerCCCount(); ++j)
      if ( (*i)->GetDist( GlobalDataBase::GetPlayerCCs()[j])
          < (*i)->GetDist( GlobalDataBase::GetPlayerCCs()[best_id] ) )
        best_id = j;

    squadAI[best_id]->AddTank( *i );
  }
}

namespace {
  void EngageVisitPlayer(const int &player_index, int &player_group, int &enemy_group, const int *graph, const int &num_player, const int &num_enemy);
  void EngageVisitEnemy(const int &enemy_index, int &player_group, int &enemy_group, const int *graph, const int &num_player, const int &num_enemy);

  void EngageVisitPlayer(const int &player_index, int &player_group, int &enemy_group, const int *graph, const int &num_player, const int &num_enemy) {
    player_group |= (1 << player_index);
    const int *offseted = graph + player_index * MAX_SQUAD_NUMBER;
    for (int i = 0; i < num_enemy; ++i)
      if ( (offseted[i] == 1) && !(enemy_group & (1 << i)))
        EngageVisitEnemy(i, player_group, enemy_group, graph, num_player, num_enemy);
  }

  void EngageVisitEnemy(const int &enemy_index, int &player_group, int &enemy_group, const int *graph, const int &num_player, const int &num_enemy) {
    enemy_group |= (1 << enemy_index);
    for (int i = 0; i < num_player; ++i)
      if ( graph[i * MAX_SQUAD_NUMBER + enemy_index] == 1
          && !(player_group & (1 << i)))
        EngageVisitPlayer(i, player_group, enemy_group, graph, num_player, num_enemy);
  }
}

bool GeneralAI::SquadMerge() {
  //merge all near squads
  if(aiMode == AI_DEFAULT) {
    // center of mass merge
    for (size_t i = 0; i < squadAI.size(); ++i) {
      if (squadAI[i]->HasTarget()) continue;
      for (size_t j = i+1; j < squadAI.size(); ++j) {
        if(squadAI[j]->HasTarget()) continue;
        if (squadAI[i]->GetCenterOfMass().SqrDist(squadAI[j]->GetCenterOfMass()) < SQUAD_MERGE_DIST ) {
          swap(squadAI[j], squadAI[squadAI.size() - 1]);
          squadAI[i]->Merge(squadAI[squadAI.size() - 1]);
          squadAI.pop_back();
          return true;
        }
      }
    }
  } else {
    // nearest squad merge
    for (size_t i = 0; i < squadAI.size(); ++i) {
      if (squadAI[i]->HasTarget()) continue;
      for (size_t j = i+1; j < squadAI.size(); ++j) {
        if(squadAI[j]->HasTarget()) continue;
        if(squadAI[i]->TestDist(squadAI[j], SQUAD_TANK_MERGE_DIST)) {
          swap(squadAI[j], squadAI[squadAI.size() - 1]);
          squadAI[i]->Merge(squadAI[squadAI.size() - 1]);
          squadAI.pop_back();
          return true;
        }
      }
    }
  }
  return false;
}

int GeneralAI::ScanEnemy(const WorldCoor& world_coor, const int& sqr_range) {
  sint4 harassSet = 0, harassTankCount = 0;
  for (EnemyTank* const* i = GlobalDataBase::GetEnemyTanks(),
      * const* ei = GlobalDataBase::GetEnemyTanksEnd(); i != ei; ++i)
    if ( world_coor.SqrDist( (*i)->GetWorldCoor() ) < sqr_range ) {
      harassSet |= (1 << (*i)->GetCluster()->GetID() );
    }
  int bk = harassSet;
  while(bk) {
    int last_bit = bk ^ (bk & (bk - 1));
    bk ^= last_bit;
    last_bit = __builtin_ctz(last_bit);
    harassTankCount += Clustering::GetClusters()[last_bit]->GetSize();
  }
  return harassTankCount;
}

void GeneralAI::StandyAI() {
  WorldCoor assembly_point;
  FOR_RANGE(cc, GlobalDataBase::GetPlayerCCs(), GlobalDataBase::GetPlayerCCsEnd()) {
    if ((*cc)->IsDead()) continue;
    assembly_point = (*cc)->GetWorldCoor();
    break;
  }
  if(squadAI.size() > 1) {
    // if we have squads elsewhere, wait until everyone merges
    // into a huge big squad
    for (int i = 0, ei = squadAI.size(); i < ei; ++i) {
      if(squadAI[i]->HasTarget()) continue;
      //squadAI[i]->SetTarget(new Target_Assemble(assembly_point));
      squadAI[i]->SetAIMode(AI_DEFENSIVE);
      squadAI[i]->SetDefaultActionWaypoint(assembly_point.GetMapCoor());
      squadAI[i]->ResetTargetCooldown();
    }
  } else {
    // nothing to do, just resume normal functionality
    aiMode = AI_DEFAULT;
  }
}

void GeneralAI::DefensiveAI() {
  WorldCoor assembly_point;
  // find our only CC
  FOR_RANGE(cc, GlobalDataBase::GetPlayerCCs(), GlobalDataBase::GetPlayerCCsEnd()) {
    if ((*cc)->IsDead()) continue;
    assembly_point = (*cc)->GetWorldCoor();
    break;
  }
  if(squadAI.size() > 1) {
    // if we have squads elsewhere, wait until everyone merges
    // into a huge big squad
    for (int i = 0, ei = squadAI.size(); i < ei; ++i) {
      if(squadAI[i]->HasTarget()) continue;
      //squadAI[i]->SetTarget(new Target_Assemble(assembly_point));
      squadAI[i]->SetAIMode(AI_DEFENSIVE);
      squadAI[i]->SetDefaultActionWaypoint(assembly_point.GetMapCoor());
      //squadAI[i]->ResetTargetCooldown();
    }
  } else {
    // find nearest enemy cluster and assign an attack function
    int best_dist = 9999999;
    if(squadAI[0]->HasTarget()) return; //can't do anything for now
    Cluster* best_cl;
    FOR_RANGE(cl,
        Clustering::GetClusters(),
        Clustering::GetClustersEnd()) {
      if( !MapBox::Walkable((*cl)->GetCenterOfMass().GetMapCoor()))
        continue;
      const int d = (*cl)->GetCenterOfMass().SqrDist(assembly_point);
      if(d < best_dist) {
        best_dist = d;
        best_cl = (*cl);
      }
    }
    if(best_dist < 9999999) {
      const WorldCoor& wp = best_cl->GetCenterOfMass();
      const int req = RescueSquadCount(
          ScanEnemy(wp, TARGET_FAR_SCAN_SQR_RANGE));
      SquadAI* sq = squadAI[0]->Split(wp, req);
      if(sq != NULL) { // erm... do nothing?
        squadAI.push_back(sq);
        sq->SetTarget(new Target_Area(wp.GetMapCoor()));
        squadAI[0]->ResetTargetCooldown();
      } else if(Clustering::GetClusterCount() == 1
          && GlobalDataBase::GetEnemyTankCount() < GlobalDataBase::GetPlayerTankCount()) {
        squadAI[0]->SetTarget(new Target_Area(wp.GetMapCoor()));
      } else {
        // do nothing!!! we DIE!!!
      }
    } else {
      // nothing to do, just resume normal functionality
      aiMode = AI_DEFAULT;
      for (int i = 0; i < squadAI.size(); i++) {
        squadAI[i]->SetAIMode(AI_DEFAULT);
      }
    }
  }
}

void GeneralAI::DefaultAI() {
  for (int i = 0, ei = squadAI.size(); i < ei; ++i) {
    if(squadAI[i]->HasTarget()) continue; //ignore this squad for now
    FOR_RANGE(cc,
        GlobalDataBase::GetPlayerCCs(),
        GlobalDataBase::GetPlayerCCsEnd()) {
      // if we are near an enemy CC, and we have >1 CC
      // then don't assign mission
      // if have multiple squad attacking enemy CC, some squad may be
      // redundant
      if ((*cc)->IsDead()) continue;
      if ((*cc)->GetRescuer() > 0) continue; //do not assign multiple team
      if ((*cc)->GetHarassSet() > 0) { 
        const int req = RescueSquadCount((*cc)->GetHarassTankCount());
        if (CalcRescueTime(*cc, squadAI[i], req) > RESCUE_TIME_OFFSET) {
          LOG(ferr << "Squad " << i << ": Rescue CC" << endl);
          WorldCoor center_of_mass = squadAI[i]->GetCenterOfMass();
          SquadAI* sq = squadAI[i]->Split((*cc)->GetWorldCoor(), req);
          if(sq != NULL) {
            LOG(ferr << " Split squad!" << endl );
            swap(squadAI[i], sq);
            squadAI.push_back(sq);
            sq->ResetTargetCooldown();
            //sq->SetTarget(new Target_Cooldown(TARGET_COOLDOWN_TIMER));
          }
          // finally
          squadAI[i]->SetTarget(new Target_PlayerCC(*cc));
          break;
        }
      }
    } 
  }
}

// Build Group
void GeneralAI::Update() {
  //Upate squads status
  LOG(ferr << "GeneralAI::Update(), before everything" << endl);
  CLEAR_DEAD(squadAI);
  //CheckBattleStatus();
  //1. Get updated enemy clustering information
  //1.1 Decide how many clusters of enemy unit
  LOG(ferr << "After clear dead squads" << endl);

  for (size_t i = 0; i < squadAI.size(); ++i) {
    //cerr << " Squad " << i << " update " << endl;
    squadAI[i]->Update();
  }

  while(SquadMerge());

  int enemy_count = GlobalDataBase::GetEnemyTankCount();
  EnemyTank** enemy_tanks = GlobalDataBase::GetEnemyTanks();

  //LOG(ferr << "Before dest assignment" << endl);

  if(GlobalDataBase::GetPlayerCCCount() == 1
      && GlobalDataBase::GetEnemyCCCount() > 1
      && (GlobalDataBase::GetPlayerTankCount() - GlobalDataBase::GetEnemyCCCount() * 0.8f) > 0) {
    aiMode = AI_DEFENSIVE;
  }

  switch(aiMode) {
    case AI_DEFENSIVE:
      DefensiveAI();
      break;
    case AI_DEFAULT:
      // GENERAL CASE
      DefaultAI();
      break;
  }
  LOG(ferr << "After dest assignment" << endl);
  // TODO: Remove this hack...

#ifdef OLD_AI
  //n. Make strategic decision
  memset(engagingMap, 0, MAX_SQUAD_NUMBER*MAX_SQUAD_NUMBER);
  for (int i = 0, ei = squadAI.size(); i < ei; i++) {
    squadAI.at(i)->EngagingEnemyCluster(engagingMap, MAX_SQUAD_NUMBER); 
  }

  ferr << "The engaging matrix is:" << endl;
  for (int i = 0, ei = squadAI.size(); i < ei; i++) {
    for (int j = 0, ej = medoidsObj.size(); j < ej; j++) {
      ferr << engagingMap[i*MAX_SQUAD_NUMBER+j] << " ";
    }
    ferr << endl;
  }
#endif

  //#ifdef TABOO_AI
  // Set taboo on where large enemy clusters locate
  if(aiMode == AI_DEFAULT) {
    MapCoor newTabooLeftTop[MAX_CLUSTER_COUNT];
    MapCoor newTabooRightBottom[MAX_CLUSTER_COUNT];
    int newTabooCount = 0, delta = 0;
    // Go through all clusters
    Cluster* const* cluster_begin = Clustering::GetClusters();
    Cluster* const* cluster_end = Clustering::GetClustersEnd();

    for (Cluster* const* i = cluster_begin; i != cluster_end; ++i) {
      Cluster* cluster = *i;
      if (cluster->GetSize() >= 15) {
        delta = (cluster->GetSize() - 10) / 4 + 5;
        MapCoor center = MapBox::GetMapCoor(cluster->GetCenterOfMass());
        newTabooLeftTop[newTabooCount]
          = MapCoor(max(center.r - delta, 0), max(center.c - delta, 0) );
        newTabooRightBottom[newTabooCount]
          = MapCoor(min(center.r + delta, 63), min(center.c + delta, 63));
        MapBox::AddTaboo(newTabooLeftTop[newTabooCount],
            newTabooRightBottom[newTabooCount]);
        ++newTabooCount;
      }
    }
    // Delete all old ones
    while(tabooCount--) MapBox::RemoveTaboo(
        clusterTabooLeftTop[tabooCount],
        clusterTabooRightBottom[tabooCount]);
    tabooCount = newTabooCount;
    for (int i = 0; i < newTabooCount; ++i) {
      clusterTabooLeftTop[i] = newTabooLeftTop[i];
      clusterTabooRightBottom[i] = newTabooRightBottom[i];
    }
  }
  //#endif

#ifdef OLD_AI

  int processed_player = 0;
  int processed_enemy = 0;
  int player_groups[MAX_SQUAD_NUMBER], enemy_groups[MAX_SQUAD_NUMBER];
  int player_group_sizes[MAX_SQUAD_NUMBER], enemy_group_sizes[MAX_SQUAD_NUMBER];
  int group_idx = 0;
  for (int i = 0, ei = squadAI.size(); i < ei; ++i) {
    // Have we dealt with this squad already?
    if (processed_player & (1 << i)) continue;
    player_groups[group_idx] = enemy_groups[group_idx] = 0;
    EngageVisitPlayer(i, player_groups[group_idx], enemy_groups[group_idx],
        engagingMap, squadAI.size(), medoidsObj.size());
    // So player_group and enemy_group are found
    processed_player |= player_groups[group_idx];
    processed_enemy |= enemy_groups[group_idx];
    // Calculate the squad sizes
    player_group_sizes[group_idx] = enemy_group_sizes[group_idx] = 0;
    for (int j = 0, ej = squadAI.size(); j < ej; ++j)
      if ( (1<<j) & player_groups[group_idx] ) // squad j in the group
        player_group_sizes[group_idx] += squadAI[j]->NumOfTanks();
    for (int j = 0, ej = medoidsObj.size(); j < ej; ++j)
      if ( (1<<j) & enemy_groups[group_idx] ) // Medoid j in the group
        enemy_group_sizes[group_idx] += medoidsSize[j];
    ++group_idx;
  }
  // Consider stray enemies
  for (int i = 0, ei = medoidsObj.size(); i < ei; ++i) {
    // Have we dealt with this squad already?
    if (processed_enemy & (1 << i)) continue;
    player_groups[group_idx] = 0;
    enemy_groups[group_idx] = (1 << i);
    //processed_enemy |= enemy_groups[group_idx]; // This is useless
    // Calculate the squad sizes
    player_group_sizes[group_idx] = 0;
    enemy_group_sizes[group_idx] = medoidsSize[i];
    ++group_idx;
  }

  // Group information and size found
  for (int i = 0; i < group_idx; i++) {
    if (!player_groups[i]) break;
    if (player_group_sizes[i] < enemy_group_sizes[i]) {
      //TODO: ask other available squads for help!!!!
      //if we have only one squad in this battle group, retreat
      if ( BIT_COUNT[player_groups[i]] == 1) {

      }
    }
  }
#endif
}

void GeneralAI::Act() {
  for (int i = 0; i < squadAI.size(); i++) {
    LOG(ferr << "Squad " << i << "::Act()" << endl);
    squadAI[i]->Act();
  }
}

