#ifndef MAP_BOX_H
#define MAP_BOX_H

#include "Primitive.H"
#include "GameTile.H"
#include "Game.H"

namespace {
  // Hard Coded now
  const int MAP_NUM_ROW = 64;
  const int MAP_NUM_COL = 64;
  const int MAP_TILE_SIZE = 16;
  const int TABOO_DURATION = 4;
}

class MapBox {
  public:
    // The MapBox is in an inconsistent state until this Init is called
    // No one should attempt to access any other functionalities until
    // this Init is called
    static void Init(const Map<GameTile>& m);

    /* Start of WorldCoor <-> MapCoor conversions */
    static inline sint4 GetLeft(const MapCoor& coor) {
      return coor.c * MAP_TILE_SIZE;
    }

    static inline sint4 GetRight(const MapCoor& coor) {
      return (coor.c+1) * MAP_TILE_SIZE;
    }

    static inline sint4 GetTop(const MapCoor& coor) {
      return coor.r * MAP_TILE_SIZE;
    }

    static inline sint4 GetBottom(const MapCoor& coor) {
      return (coor.r+1) * MAP_TILE_SIZE;
    }

    static inline sint4 GetCenterX(const MapCoor& coor) {
      return coor.c * MAP_TILE_SIZE + MAP_TILE_SIZE / 2;
    }

    static inline sint4 GetCenterY(const MapCoor& coor) {
      return coor.r * MAP_TILE_SIZE + MAP_TILE_SIZE / 2;
    }

    static inline WorldCoor GetCenter(const MapCoor& coor) {
      return WorldCoor(GetCenterX(coor), GetCenterY(coor));
    }

    static inline WorldCoor GetLeftTop(const MapCoor& coor) {
      return WorldCoor(GetLeft(coor), GetTop(coor));
    }

    static inline WorldCoor GetRightBottom(const MapCoor& coor) {
      return WorldCoor(GetRight(coor), GetBottom(coor));
    }

    static inline WorldCoor GetLeftBottom(const MapCoor& coor) {
      return WorldCoor(GetLeft(coor), GetBottom(coor));
    }

    static inline WorldCoor GetRightTop(const MapCoor& coor) {
      return WorldCoor(GetRight(coor), GetTop(coor));
    }

    static inline sint4 GetRow(const WorldCoor& coor) {
      //return coor.y / MAP_TILE_SIZE;
      return coor.y >> 4; // CAREFUL! Because MAP_TILE_SIZE = 16
    }

    static inline sint4 GetCol(const WorldCoor& coor) {
      // return coor.x / MAP_TILE_SIZE;
      return coor.x >> 4; // CAREFUL!
    }

    static inline MapCoor GetMapCoor(const WorldCoor& coor) {
      return MapCoor( GetRow(coor), GetCol(coor) );
    }

    static inline MapCoor GetMapCoor(const sint4& x, const sint4& y) {
      return MapCoor( y / MAP_TILE_SIZE, x / MAP_TILE_SIZE );
    }
    /* End of WorldCoor <-> MapCoor conversions */


    // Turn a map coordinate into an ID
    static inline sint4 CoorToID(const MapCoor& coor) {
      return coor.r * MAP_NUM_COL + coor.c;
    }

    static inline sint4 RCToID(const sint4& r, const sint4& c) {
      return r * MAP_NUM_COL + c;
    }

    static inline MapCoor IDToCoor(const sint4 &id) {
      //return MapCoor(id / MAP_NUM_COL, id % MAP_NUM_COL);
      // CAREFUL! MAP_NUM_COL = 64
      return MapCoor(id >> 6, id & (MAP_NUM_COL - 1) );
    }

    static inline bool InMap(const sint4& r, const sint4& c) {
      return r >= 0 && r < MAP_NUM_ROW && c >= 0 && c < MAP_NUM_COL;
    }

    static inline bool InMap(const MapCoor& coor) {
      return InMap(coor.r, coor.c);
    }

    static inline bool HasTaboo(const MapCoor& coor) {
      return taboo[coor.r][coor.c];
    }

    // Check if this tile is accessible or not
    static inline bool Walkable(const MapCoor& coor) {
      return InMap(coor)
        && (!inaccessible[coor.r][coor.c])
        && (!taboo[coor.r][coor.c]);
    }

    static inline bool Accessible(const MapCoor& coor) {
      return InMap(coor) && (!inaccessible[coor.r][coor.c]);
    }

    // Please note that the two map coordinates are both inclusive
    
    // GENERAL TABOO
    static inline void AddTaboo(const MapCoor& left_top,
        const MapCoor& right_bottom) {
      for (int r = left_top.r; r <= right_bottom.r; ++r)
        for (int c = left_top.c; c <= right_bottom.c; ++c)
          ++taboo[r][c];
    }

    static inline void RemoveTaboo(const MapCoor& left_top,
        const MapCoor& right_bottom) {
      for (int r = left_top.r; r <= right_bottom.r; ++r)
        for (int c = left_top.c; c <= right_bottom.c; ++c)
          --taboo[r][c];
    }

    static inline void AddTaboo(const MapCoor& coor) {
      ++taboo[coor.r][coor.c];
    }

    static inline void RemoveTaboo(const MapCoor& coor) {
      --taboo[coor.r][coor.c];
    }

    // TERRAIN TABOO
    static inline void AddTerrain(const MapCoor& left_top,
        const MapCoor& right_bottom) {
      for (int r = left_top.r; r <= right_bottom.r; ++r)
        for (int c = left_top.c; c <= right_bottom.c; ++c)
          ++inaccessible[r][c];
    }

    static inline void RemoveTerrain(const MapCoor& left_top,
        const MapCoor& right_bottom) {
      for (int r = left_top.r; r <= right_bottom.r; ++r)
        for (int c = left_top.c; c <= right_bottom.c; ++c)
          --inaccessible[r][c];
    }

    static inline void AddTerrain(const MapCoor& coor) {
      ++inaccessible[coor.r][coor.c];
    }

    static inline void RemoveTerrain(const MapCoor& coor) {
      --inaccessible[coor.r][coor.c];
    }

  private:
    // A bitmap of the current map, 0 ==> walkable
    static char taboo[MAP_NUM_ROW][MAP_NUM_COL];
    static char inaccessible[MAP_NUM_ROW][MAP_NUM_COL];
    static int StaticInit();
    static const int dummy;
};

#endif
