#include "PathFinder.H"
#include "DebugHeader.H"

/* 
   Copyright 2007 Man Hon Chan, Zhangbo Liu, Simon Suyadi
   This code is Licensed under GPL.
 */

DEF_FERR("PathFinder");

namespace {
  const MapDirection ALL_DIR[] = {
    ORTHOGONAL_DIR[0],
    ORTHOGONAL_DIR[1],
    ORTHOGONAL_DIR[2],
    ORTHOGONAL_DIR[3],
    DIAGONAL_DIR[0],
    DIAGONAL_DIR[1],
    DIAGONAL_DIR[2],
    DIAGONAL_DIR[3],
  };

}

/* These attributes are SHARED, so don't multithread this class */
sint4 PathFinder::q[MAX_EDGE_COST][MAP_NUM_ROW * MAP_NUM_COL];
int PathFinder::indices[MAP_NUM_ROW * MAP_NUM_COL];
int PathFinder::qSize[MAX_EDGE_COST];

const sint4* PathFinder::GetDist(const MapCoor& coor) {
  // First, check if we have calculated this before
  int id = MapBox::CoorToID(coor);
  //if (!Walkable(coor)) return NULL; // It is really not walkable
  if(!MapBox::Accessible(coor)) return NULL; //impossible

  if (destToCache[id] != -1) {
    int idx = destToCache[id];
    timestamps[idx] = currTime++;
    return cachedMaps[ idx ];
  }

  // It is not cached!
  // Find one that is good
  int to_discard = 0;
  for (int i = 1; i < MAX_DIST_CACHE_SIZE; ++i)
    if (timestamps[i] < timestamps[to_discard])
      to_discard = i;

  // I can now put stuff in to_discard
  destToCache[ id ] = to_discard;
  // It is no longer valid
  destToCache[cacheToDest[to_discard]] = -1;
  cacheToDest[to_discard] = id;

  // Build the BFS using cachedMaps[to_discard]
  Dijkstra(cachedMaps[to_discard], coor);

  return cachedMaps[to_discard];
}

void PathFinder::Dijkstra(sint4* dist_matrix, const MapCoor& coor) {
  //if unwalkable... then erm... ignore?
  //if (!MapBox::Walkable(coor)) return;

  // Fill the map with not found
  memset(dist_matrix, 0xff, sizeof(sint4) * MAP_NUM_ROW * MAP_NUM_COL );
  memset(qSize, 0, sizeof(qSize));

  MapCoor curr = coor, next;
  sint4 idx, dist = 0, empty_count = 0, curr_id = 0, next_id = 0;
  idx = qSize[0]++;
  next_id = MapBox::CoorToID(coor);
  curr_id = MapBox::CoorToID(curr);
  dist_matrix[next_id] = 0;
  indices[next_id] = idx;
  q[0][idx] = curr_id; // The current

  //LOG( ferr << "dijkstra at=" << coor.r << "," << coor.c << endl );
  for (;empty_count < MAX_EDGE_COST; ++dist) {
    int dist_ofs = dist & (MAX_EDGE_COST - 1);
    if (qSize[dist_ofs] == 0) {
      ++empty_count;
      continue;
    }
    empty_count = 0;
    //LOG( ferr << " dist=" << dist << " off=" << dist_ofs << endl );
    while(qSize[dist_ofs] > 0) {
      curr_id = q[dist_ofs][--qSize[dist_ofs]];
      curr = MapBox::IDToCoor(curr_id);
      //LOG( ferr << " curr=" << curr.r << "," << curr.c << endl );
      // Try all my neighbours
      for (int s = 0; s < 8; ++s) {
        const MapDirection &dir = ALL_DIR[s];
        //if (!CanGo(curr, dir)) continue;
        if(!CanAcess(curr, dir)) continue; //impassable
        // If dir is diagonal, dir.dr * dir.dr + dir.dc * dir.dc = 2
        // This is to avoid branching
        int new_dist = dist + 2 + 1 * (dir.dr * dir.dr + dir.dc * dir.dc);
        if(!CanGo(curr, dir)) new_dist += 20; //taboo here, add cost
        next = curr + dir;
        next_id = MapBox::CoorToID(next);
        if (/*dist_matrix[next_id] != -1 &&*/ dist_matrix[next_id] > new_dist) {
          int d = dist_matrix[next_id] & (MAX_EDGE_COST - 1);
          q[d][indices[next_id]] = q[d][--qSize[d]];
          dist_matrix[next_id] = -1;
        }
        if (dist_matrix[next_id] == -1) {
          int d = new_dist & (MAX_EDGE_COST - 1);
          idx = qSize[d]++;
          q[d][idx] = next_id;
          indices[next_id] = idx;
          dist_matrix[next_id] = new_dist;
        }
      }
    }
  } // End of empty_count loop
}

