#ifndef PATH_FINDER_H
#define PATH_FINDER_H

#include "MapBox.H"
#include "GlobalDataBase.H"

namespace {
  // How many different places we want to store?
  const int MAX_DIST_CACHE_SIZE = 16;
  const int MAX_EDGE_COST = 32;
}

class PathFinder {
  public:

    /* Starting instance interface */
    PathFinder() {
      Reset();
    }

    inline void Reset() {
      currTime = 0;
      memset(tabooExpiry, -1, sizeof(tabooExpiry));
      memset(destToCache, -1, sizeof(destToCache) );

      // All timestamps are invalid to begin with
      memset(timestamps, -1, sizeof(timestamps));

      // No cache is matched to anything
      memset(cacheToDest, -1, sizeof(cacheToDest));
    }

    inline void SetTaboo(const MapCoor& coor) {
      tabooExpiry[coor.r][coor.c]
        = GlobalDataBase::GetGameTick() + TABOO_DURATION;
    }

    inline bool Walkable(const MapCoor& coor) const {
      return MapBox::Walkable(coor)
        && (tabooExpiry[coor.r][coor.c] < GlobalDataBase::GetGameTick() );
    }

    // Test whether we can go from this coordinate along a direction.
    // The direction must be one of the eight orthogonal direction
    inline bool CanGo(const MapCoor& coor, const MapDirection &dir) const {
      sint4 size = dir.dr * dir.dr + dir.dc * dir.dc;
      if (size == 1)
        return Walkable(coor + dir);
      else {
        return Walkable(coor + dir)
          && Walkable( MapCoor(coor.r + dir.dr, coor.c) )
          && Walkable( MapCoor(coor.r, coor.c + dir.dc) ) ;
      }
    }

    inline bool CanAcess(const MapCoor& coor, const MapDirection &dir) const {
      sint4 size = dir.dr * dir.dr + dir.dc * dir.dc;
      if (size == 1)
        return MapBox::Accessible(coor + dir);
      else {
        return MapBox::Accessible(coor + dir)
          && MapBox::Accessible( MapCoor(coor.r + dir.dr, coor.c) )
          && MapBox::Accessible( MapCoor(coor.r, coor.c + dir.dc) ) ;
      }
    }

    // Get the distance from ALL location to the supplied destination square (r,c)
    // The returned array will be expressed linearly.
    // Get the (x,y) element at y * width + x
    const sint4* GetDist(const MapCoor& coor);

  private:
    /* These can be reused */
    static sint4 q[MAX_EDGE_COST][MAP_NUM_ROW * MAP_NUM_COL];
    static int indices[MAP_NUM_ROW * MAP_NUM_COL];
    static int qSize[MAX_EDGE_COST];

    //static int best[MAX_NUM_ROW][MAP_NUM_COL];
    /* End of static reuse */

    //static int cost[MAP_NUM_ROW][MAP_NUM_COL];

    // This stores the expiry date
    sint4 tabooExpiry[MAP_NUM_ROW][MAP_NUM_COL];
    sint4 currTime; // Current time as in game_tick

    // Store all cached destination
    sint4 cachedMaps[MAX_DIST_CACHE_SIZE][MAP_NUM_ROW * MAP_NUM_COL];

    char destToCache[MAP_NUM_ROW * MAP_NUM_COL];
    sint4 cacheToDest[MAX_DIST_CACHE_SIZE];
    int timestamps[MAX_DIST_CACHE_SIZE];

    void Dijkstra(sint4* dist_matrix, const MapCoor& coor);
};

#endif
