#include "MapBox.H"
#include <string>

namespace {
  const MapDirection ALL_DIR[] = {
    ORTHOGONAL_DIR[0],
    ORTHOGONAL_DIR[1],
    ORTHOGONAL_DIR[2],
    ORTHOGONAL_DIR[3],
    DIAGONAL_DIR[0],
    DIAGONAL_DIR[1],
    DIAGONAL_DIR[2],
    DIAGONAL_DIR[3],
  };

  bool TileWalkable(const GameTile& tile) {
    sint4 hNW, hNE, hSE, hSW;
    Tile::Split split;
    tile.get_topo(hNW, hNE, hSE, hSW, split);
    return (split == Tile::NO_SPLIT) && !(hNW > 0 || hNE > 0 || hSE > 0 || hSW > 0);
  }

}

char MapBox::bitmap[MAP_NUM_ROW][MAP_NUM_COL];

// Store all cached destination
sint4 MapBox::cachedMaps[MAX_DIST_CACHE_SIZE][MAP_NUM_ROW * MAP_NUM_COL];

char MapBox::destToCache[MAP_NUM_ROW * MAP_NUM_COL];
sint4 MapBox::cacheToDest[MAX_DIST_CACHE_SIZE];

int MapBox::timestamps[MAX_DIST_CACHE_SIZE];
int MapBox::currTime;

sint4 MapBox::q[MAX_EDGE_COST][MAP_NUM_ROW * MAP_NUM_COL];
int MapBox::indices[MAP_NUM_ROW * MAP_NUM_COL];
int MapBox::qSize[MAX_EDGE_COST];

const int MapBox::dummy = StaticInit();

int MapBox::StaticInit() {
  return 0;
}

void MapBox::Init(const Map<GameTile>& m) {

  // Fill in the entries
  for (int r = 0; r < MAP_NUM_ROW; ++r)
    for (int c = 0; c < MAP_NUM_COL; ++c) {
      const GameTile &t = m(c, r); // Convert x,y into r,c.....
      bool accessible = TileWalkable(t);
      bitmap[r][c] = accessible?0:1;
    }

  currTime = 0;
  memset(destToCache, -1, sizeof(destToCache) );

  // All timestamps are invalid to begin with
  memset(timestamps, -1, sizeof(timestamps));

  // No cache is matched to anything
  memset(cacheToDest, -1, sizeof(cacheToDest));

}

const sint4* MapBox::GetDist(const MapCoor& coor) {
  // First, check if we have calculated this before
  int id = CoorToID(coor);
  if (!Walkable(coor)) return NULL; // It is really not walkable

  if (destToCache[id] != -1) {
    int idx = destToCache[id];
    timestamps[idx] = currTime++;
    return cachedMaps[ idx ];
  }

  // It is not cached!
  // Find one that is good
  int to_discard = 0;
  for (int i = 1; i < MAX_DIST_CACHE_SIZE; ++i)
    if (timestamps[i] < timestamps[to_discard])
      to_discard = i;

  // I can now put stuff in to_discard
  destToCache[ id ] = to_discard;
  // It is no longer valid
  destToCache[cacheToDest[to_discard]] = -1;
  cacheToDest[to_discard] = id;

  // Build the BFS using cachedMaps[to_discard]
  Dijkstra(cachedMaps[to_discard], coor);

  return cachedMaps[to_discard];
}

void MapBox::Dijkstra(sint4* dist_matrix, const MapCoor& coor) {
  // Fill the map with not found
  memset(dist_matrix, 0xff, sizeof(sint4) * MAP_NUM_ROW * MAP_NUM_COL );
  memset(qSize, 0, sizeof(qSize));

  MapCoor curr = coor, next;
  sint4 idx, dist = 0, empty_count = 0, curr_id = 0, next_id = 0;
  idx = qSize[0]++;
  next_id = CoorToID(coor);
  curr_id = CoorToID(curr);
  dist_matrix[next_id] = 0;
  indices[next_id] = idx;
  q[0][idx] = curr_id; // The current

  for (;empty_count < 8; ++dist) {
    int dist_ofs = dist & 7;
    if (qSize[dist_ofs] == 0) {
      ++empty_count;
      continue;
    }
    empty_count = 0;
    while(qSize[dist_ofs] > 0) {
      curr_id = q[dist_ofs][--qSize[dist_ofs]];
      curr = IDToCoor(curr_id);
      // Try all my neighbours
      for (int s = 0; s < 8; ++s) {
        MapDirection dir = ALL_DIR[s];
        if (!CanGo(curr, dir)) continue;
        // If dir is diagonal, dir.dr * dir.dr + dir.dc * dir.dc = 2
        // This is to avoid branching
        int new_dist = dist + 3 + 2 * (dir.dr * dir.dr + dir.dc * dir.dc);
        next = curr + dir;
        next_id = CoorToID(next);
        if (/*dist_matrix[next_id] != -1 &&*/ dist_matrix[next_id] > new_dist) {
          int d = dist_matrix[next_id] & 7;
          q[d][indices[next_id]] = q[d][--qSize[d]];
          dist_matrix[next_id] = -1;
        }
        if (dist_matrix[next_id] == -1) {
          int d = new_dist & 7;
          idx = qSize[d]++;
          q[d][idx] = next_id;
          indices[next_id] = idx;
          dist_matrix[next_id] = new_dist;
        }
      }
    }
  } // End of empty_count loop

}

