#include "DebugHeader.H"
#include "GlobalDataBase.H"
#include "Random.H"
#include "MapBox.H"
#include "FFEngine.H"
#include "FFConst.H"

#include "GameObjWrapper.H"
#include "PlayerTank.H"
#include "EnemyObj.H"
#include "NeutralObj.H"

#include "FFPrimitive.H"
#include "FFData.H"

#include <algorithm>
#include <cmath>
#include "Game.H"

#include "CollisionEngine.H"
#include "XTerrainEngine.H"

DEF_FERR("FFEngine");

namespace {
  const MapDirection ALL_DIR[] = {
    ORTHOGONAL_DIR[0],
    ORTHOGONAL_DIR[1],
    ORTHOGONAL_DIR[2],
    ORTHOGONAL_DIR[3],
    DIAGONAL_DIR[0],
    DIAGONAL_DIR[1],
    DIAGONAL_DIR[2],
    DIAGONAL_DIR[3],
  };

  const sint4 INF_DIST = 0x3f3f3f3f;

  CollisionEngine collisionEngine;
}

// Declare all variables
FFPlayerData** FFEngine::playerBucketsEnd[NUM_MAP_ROW][NUM_MAP_COL];
FFPlayerData* FFEngine::playerBuckets[NUM_MAP_ROW][NUM_MAP_COL][PLAYER_BUCKET_SIZE];
FFPlayerData** FFEngine::predictedBucketsEnd[NUM_MAP_ROW][NUM_MAP_COL];
FFPlayerData* FFEngine::predictedBuckets[NUM_MAP_ROW][NUM_MAP_COL][PREDICTED_BUCKET_SIZE];
FFEnemyData** FFEngine::enemyBucketsEnd[NUM_MAP_ROW][NUM_MAP_COL];
FFEnemyData* FFEngine::enemyBuckets[NUM_MAP_ROW][NUM_MAP_COL][ENEMY_BUCKET_SIZE];
FFNeutralData** FFEngine::neutralBucketsEnd[NUM_MAP_ROW][NUM_MAP_COL];
FFNeutralData* FFEngine::neutralBuckets[NUM_MAP_ROW][NUM_MAP_COL][NEUTRAL_BUCKET_SIZE];
FFPlayerData* FFEngine::processQueue[MAX_NUM_TANK];
FFPlayerData** FFEngine::processQueueEnd;


#define FOR_RANGE(it, begin, end) for ( typeof(begin) it = (begin),\
    _end_of_loop_indicator_ = (end); it != _end_of_loop_indicator_; ++it)
#define FOR_RANGE2(it, begin, end) for ( typeof(begin) it = (begin),\
    _end_of_loop_indicator2_ = (end); it != _end_of_loop_indicator2_; ++it)

// Must be called after the GlobalDataBase has been "constructed"
// That means all relevant static init are done
void FFEngine::Install() {
  for (int r = 0; r < NUM_MAP_ROW; ++r)
    for (int c = 0; c < NUM_MAP_COL; ++c) {
      predictedBucketsEnd[r][c] = &predictedBuckets[r][c][0];
      neutralBucketsEnd[r][c] = &neutralBuckets[r][c][0];
      playerBucketsEnd[r][c] = &playerBuckets[r][c][0];
      enemyBucketsEnd[r][c] = &enemyBuckets[r][c][0];
    }

  FOR_RANGE(it, GlobalDataBase::GetEnemyTanks(),
      GlobalDataBase::GetEnemyTanksEnd())
    (*it)->ffData = new FFEnemyData( *it );

  FOR_RANGE(it, GlobalDataBase::GetPlayerTanks(),
      GlobalDataBase::GetPlayerTanksEnd())
    (*it)->ffData = new FFPlayerData( *it );

  FOR_RANGE(it, GlobalDataBase::GetNeutralObjs(),
      GlobalDataBase::GetNeutralObjsEnd())
    (*it)->ffData = new FFNeutralData( *it );

  FOR_RANGE(it, GlobalDataBase::GetPlayerCCs(),
      GlobalDataBase::GetPlayerCCsEnd())
    (*it)->ffData = new FFStaticData( *it );

  processQueueEnd = processQueue;
}

// Must be called after every thing in the GlobalDataBase has been inited
void FFEngine::Init() {
  FOR_RANGE(it, GlobalDataBase::GetEnemyTanks(),
      GlobalDataBase::GetEnemyTanksEnd()) {
    EnemyTank* tank = *it;
    AddEnemyData(tank->GetMapCoor(), (FFEnemyData*) tank->ffData);
  }

  FOR_RANGE(it, GlobalDataBase::GetPlayerTanks(),
      GlobalDataBase::GetPlayerTanksEnd()) {
    PlayerTank* tank = *it;
    AddPlayerData(tank->GetMapCoor(), (FFPlayerData*) tank->ffData);
    AddPredictedData(tank->GetWorldCoor(), (FFPlayerData*) tank->ffData);
  }

  FOR_RANGE(it, GlobalDataBase::GetNeutralObjs(),
      GlobalDataBase::GetNeutralObjsEnd()) {
    NeutralObj* sheep = *it;
    AddNeutralData(sheep->GetMapCoor(), (FFNeutralData*) sheep->ffData);
  }

  FOR_RANGE(it, GlobalDataBase::GetPlayerCCs(),
      GlobalDataBase::GetPlayerCCsEnd())
    ((FFStaticData*) (*it)->ffData)->Init();
}


void FFEngine::ClearPlayerData(const FFPlayerData* const &data) {
  RemovePlayerData(data);
  RemovePredictedData(data);
}

void FFEngine::ClearNeutralData(const FFNeutralData* const &data) {
  RemoveNeutralData(data);
}

void FFEngine::ClearEnemyData(const FFEnemyData* const &data) {
  RemoveEnemyData(data);
}



inline void FFEngine::AddPlayerData(const MapCoor& coor, FFPlayerData* data) {
#ifdef DEBUG
  if (playerBucketsEnd[coor.r][coor.c]
      - &playerBuckets[coor.r][coor.c][0] >= PLAYER_BUCKET_SIZE)
    LOG(ferr << "Adding too many players into a single bucket" << endl);
#endif
  *(playerBucketsEnd[coor.r][coor.c]++) = data;
  data->prevMapCoor = coor;
}

inline void FFEngine::AddPredictedData(const WorldCoor& coor, FFPlayerData* data) {
  const MapCoor& map_coor = MapBox::GetMapCoor(coor);
#ifdef DEBUG
  if (predictedBucketsEnd[map_coor.r][map_coor.c]
      - &predictedBuckets[map_coor.r][map_coor.c][0] >= PREDICTED_BUCKET_SIZE)
    LOG(ferr << "Adding too many players into a predicted bucket: " << coor << endl);
#endif
  *(predictedBucketsEnd[map_coor.r][map_coor.c]++) = data;
  data->predictedWorldCoor = coor;
  data->predictedMapCoor = map_coor;
}

inline void FFEngine::AddEnemyData(const MapCoor& coor, FFEnemyData* data) {
#ifdef DEBUG
  if (enemyBucketsEnd[coor.r][coor.c]
      - &enemyBuckets[coor.r][coor.c][0] >= ENEMY_BUCKET_SIZE)
    LOG(ferr << "Adding too many enemies into a single bucket" << endl);
#endif
  *(enemyBucketsEnd[coor.r][coor.c]++) = data;
  data->prevMapCoor = coor;
}

inline void FFEngine::AddNeutralData(const MapCoor& coor, FFNeutralData* data) {
#ifdef DEBUG
  if (neutralBucketsEnd[coor.r][coor.c]
      - &neutralBuckets[coor.r][coor.c][0] >= NEUTRAL_BUCKET_SIZE)
    LOG(ferr << "Adding too many sheep into a single bucket" << endl);
#endif
  *(neutralBucketsEnd[coor.r][coor.c]++) = data;
  data->prevMapCoor = coor;
}

// Will these work?
inline void FFEngine::RemovePlayerData(const FFPlayerData* const &data) {
  const MapCoor& coor = data->prevMapCoor;
  FFPlayerData** list = &playerBuckets[coor.r][coor.c][0];
  // CAREFUL! This is to speed up looping stuff....
  if ( (*list) == data )
    *list = *(--playerBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--playerBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--playerBucketsEnd[coor.r][coor.c]);
  else
    --playerBucketsEnd[coor.r][coor.c];
}

inline void FFEngine::RemovePredictedData(const FFPlayerData* const &data) {
  const MapCoor& coor = data->predictedMapCoor;
  FFPlayerData** list = &predictedBuckets[coor.r][coor.c][0];
  // CAREFUL! This is to speed up looping stuff....
  if ( (*list) == data )
    *list = *(--predictedBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--predictedBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--predictedBucketsEnd[coor.r][coor.c]);
  else
    --predictedBucketsEnd[coor.r][coor.c];
}

inline void FFEngine::RemoveEnemyData(const FFEnemyData* const &data) {
  const MapCoor& coor = data->prevMapCoor;
  FFEnemyData** list = &enemyBuckets[coor.r][coor.c][0];
  // CAREFUL! This is to speed up looping stuff....
  if ( (*list) == data )
    *list = *(--enemyBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--enemyBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--enemyBucketsEnd[coor.r][coor.c]);
  else
    --enemyBucketsEnd[coor.r][coor.c];
}

inline void FFEngine::RemoveNeutralData(const FFNeutralData* const &data) {
  const MapCoor& coor = data->prevMapCoor;
  FFNeutralData** list = &neutralBuckets[coor.r][coor.c][0];
  // CAREFUL! This is to speed up looping stuff....
  if ( (*list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else if ( *(++list) == data )
    *list = *(--neutralBucketsEnd[coor.r][coor.c]);
  else
    --neutralBucketsEnd[coor.r][coor.c];
}

void FFEngine::RebucketPlayer(PlayerTank* tank) {
  FFPlayerData* data = (FFPlayerData*) tank->ffData;
  const MapCoor& newCoor = tank->GetMapCoor();
  RemovePredictedData(data);
  if (newCoor != data->prevMapCoor) {
    RemovePlayerData(data);
    AddPlayerData(newCoor, data);
  }
}

void FFEngine::RebucketEnemy(EnemyTank* tank) {
  FFEnemyData* data = (FFEnemyData*) tank->ffData;
  const MapCoor& newCoor = tank->GetMapCoor();
  if (newCoor != data->prevMapCoor) {
    RemoveEnemyData(data);
    AddEnemyData(newCoor, data);
  }
}

void FFEngine::RebucketNeutral(NeutralObj* sheep) {
  FFNeutralData* data = (FFNeutralData*) sheep->ffData;
  const MapCoor& newCoor = sheep->GetMapCoor();
  if (newCoor != data->prevMapCoor) {
    RemoveNeutralData(data);
    AddNeutralData(newCoor, data);
  }
}

void FFEngine::UpdatePlayerData(PlayerTank* tank) {
  FFPlayerData* data = (FFPlayerData*) tank->ffData;
  const MapCoor& dest = tank->GetDest();
  const WorldCoor& dest_center = MapBox::GetCenter(dest);
  const WorldCoor& world_coor = tank->GetWorldCoor();
  const MapCoor& curr_coor = tank->GetMapCoor();
  const sint4 curr_id = MapBox::CoorToID( curr_coor );
  const sint4* dist_to_dest = MapBox::GetDist( dest );

  data->sqrEuclidDist = tank->GetWorldCoor().SqrDist(dest_center);
  data->manDist = tank->GetWorldCoor().ManDist(dest_center);
  if (dist_to_dest != NULL && dist_to_dest[curr_id] != -1) {
    data->estimatedMapDist = dist_to_dest[curr_id];
    data->hasRoute = true;
  } else {
    data->estimatedMapDist = INF_DIST;
    data->hasRoute = false;
  }

  if (dist_to_dest != NULL && dist_to_dest[curr_id] != -1) {
    sint4 curr_estimate = dist_to_dest[curr_id];
    if (curr_estimate < 5) { // So I am already there....
      data->hasRoute = false;
    } else {

      sint4 min_map_dist = 0x3f3f3f3f, min_man_dist = 0x3f3f3f3f;
      MapCoor neigh_coor;
      sint4 id, tmp_dist, tmp_man_dist;
      WorldCoor cell_center;

      for (int s = 0; s < 8; ++s) {
        MapDirection dir = ALL_DIR[s];
        if (!MapBox::CanGo(curr_coor, dir) ) continue;
        neigh_coor = curr_coor + dir;
        id = MapBox::CoorToID(neigh_coor);
        if (dist_to_dest[id] == -1) continue; // Can't go from that point
        tmp_man_dist = abs(neigh_coor.r - dest.r)
          + abs(neigh_coor.c - dest.c);
        tmp_dist = dist_to_dest[id];
        if (tmp_dist < min_map_dist
            || (tmp_dist == min_map_dist && tmp_man_dist < min_man_dist) ) {
          min_map_dist = tmp_dist;
          min_man_dist = tmp_man_dist;
          // Try to move in this direction
          cell_center = MapBox::GetCenter(neigh_coor);
          data->dx = cell_center.x - world_coor.x;
          data->dy = cell_center.y - world_coor.y;
        }

        if (min_map_dist >= 0x3f3f3f3f) {
          // I have no direction!
          data->hasRoute = false;
        }

      }

    }

  }

}

void FFEngine::Run() {

  LOG(ferr << "Begin FFEngine::Run()" << endl);

  processQueueEnd = processQueue;

  FOR_RANGE(it, GlobalDataBase::GetPlayerTanks(),
      GlobalDataBase::GetPlayerTanksEnd() ) {
    RebucketPlayer(*it);
    // Need to find out where they should move
    // and how far they are from the destination
    UpdatePlayerData( *it );
    *(processQueueEnd++) = (FFPlayerData*) (*it)->ffData;
  }

  FOR_RANGE(it, GlobalDataBase::GetEnemyTanks(),
      GlobalDataBase::GetEnemyTanksEnd() )
    RebucketEnemy(*it);

  FOR_RANGE(it, GlobalDataBase::GetNeutralObjs(),
      GlobalDataBase::GetNeutralObjsEnd() )
    RebucketNeutral( *it );

  std::sort(processQueue, processQueueEnd);

  /*
     FOR_RANGE(it, GlobalDataBase::GetPlayerTanks(),
     GlobalDataBase::GetPlayerTanksEnd() ) {
     AddPredictedData( (*it)->GetWorldCoor(), (FFPlayerData*)(*it)->ffData );
     }

     RemovePredictedData( (FFPlayerData*)
     (*GlobalDataBase::GetPlayerTanks())->ffData );

     processQueue[0] = (FFPlayerData*)
     (*GlobalDataBase::GetPlayerTanks())->ffData;

     processQueueEnd = processQueue + 1;
   */

  Random* random = GlobalDataBase::GetRandom();
  bool show_msg;

  FOR_RANGE(it, &processQueue[0], processQueueEnd) {
    //LOG(ferr << "Another tank" << endl);

    FFPlayerData* data = *it;
    PlayerTank* tank = (PlayerTank*) data->gameObjWrapper;
    const MapCoor& tank_map_coor = tank->GetMapCoor();
    const WorldCoor& tank_world_coor = tank->GetWorldCoor();
    WorldCoor cell_coor = tank_world_coor.Offset(
        MapBox::GetLeftTop(tank_map_coor) );


    //show_msg = (tank == *(GlobalDataBase::GetPlayerTanks()));
    show_msg = false;

    // Reset the engine
    collisionEngine.Reset();

    // Loop through all neighbouring terrain
    for (sint4 dr = -1; dr <= 1; ++dr)
      for (sint4 dc = -1; dc <= 1; ++dc) {
        sint4 r = tank_map_coor.r + dr;
        sint4 c = tank_map_coor.c + dc;

        collisionEngine.AddTerrain( dr, dc,
            XTerrainEngine::Get(r,c), cell_coor);
        /*
           collisionEngine.AddTerrain( dc+1, dr+1,
           XTerrainEngine::Get(r,c), cell_coor.x, cell_coor.y);
         */
      }

    if(collisionEngine.Isstuck()) show_msg = true;

    /*
       if (show_msg) {
       LOG(ferr << "My destination is " << tank->GetDest()
       << " = " << MapBox::GetCenter(tank->GetDest()) << endl);
       LOG(ferr << "My current location is " << tank->GetWorldCoor() << endl);
       LOG(ferr << "I am trying to go in the direction "
       << WorldCoor(data->dx, data->dy) << endl);
       LOG( collisionEngine.Debug(ferr) );
       LOG( ferr << " Explore the terrain: " << endl );
       for (sint4 dr = -1; dr <= 1; ++dr)
       for (sint4 dc = -1; dc <= 1; ++dc) {
       sint4 r = tank_map_coor.r + dr;
       sint4 c = tank_map_coor.c + dc;
       LOG( ferr << " At i=" << (dc+1) << " j=" << (dr+1) << " k=" << XTerrainEngine::Get(r, c) << endl );
       }
       }
     */


    /* TODO: Add buildings here */

    // Loop through all neighbouring neutral objects
    sint4 r1 = (tank_map_coor.r - 1) >? 0,
          r2 = (tank_map_coor.r + 1) <? (NUM_MAP_ROW-1),
          c1 = (tank_map_coor.c - 1) >? 0,
          c2 = (tank_map_coor.c + 1) <? (NUM_MAP_COL-1);

    for (sint4 r = r1; r <= r2; ++r)
      for (sint4 c = c1; c <= c2; ++c) {
        FOR_RANGE2(it2, &neutralBuckets[r][c][0], neutralBucketsEnd[r][c]) {
          NeutralObj* sheep = (NeutralObj*) (*it2)->gameObjWrapper;
          const WorldCoor &sheep_coor = sheep->GetWorldCoor();
          collisionEngine.AddSheep( sheep_coor.x - tank_world_coor.x,
              sheep_coor.y - tank_world_coor.y );
        }

        /* Should I add my enemies? */
        FOR_RANGE2(it2, &enemyBuckets[r][c][0], enemyBucketsEnd[r][c]) {
          EnemyTank* obstacle = (EnemyTank*) (*it2)->gameObjWrapper;
          const WorldCoor &obstacle_coor = obstacle->GetWorldCoor();
          collisionEngine.AddTank( obstacle_coor.x - tank_world_coor.x,
              obstacle_coor.y - tank_world_coor.y );
        }

        FOR_RANGE2(it2, &playerBuckets[r][c][0], playerBucketsEnd[r][c]) {
          PlayerTank* obstacle = (PlayerTank*) (*it2)->gameObjWrapper;
          if (obstacle == tank) continue;
          const WorldCoor &obstacle_coor = obstacle->GetWorldCoor();
          collisionEngine.AddTank( obstacle_coor.x - tank_world_coor.x,
              obstacle_coor.y - tank_world_coor.y );
        }

        // Future ones as well
        /*
           FOR_RANGE2(it2, &predictedBuckets[r][c][0], predictedBucketsEnd[r][c]) {
           const WorldCoor &obstacle_coor = (*it)->predictedWorldCoor;
           collisionEngine.AddTank( obstacle_coor.x - tank_world_coor.x,
           obstacle_coor.y - tank_world_coor.y );
           }
         */
        // End of adding all obstacles

      }

    int a, s; // Angle and speed/displacement
    double real_a; // The one from Dijkstra
    {
      if (data->hasRoute) {
        real_a = atan2(data->dy, data->dx);
        if (real_a < 0) real_a += PI2;
        if (show_msg) {
          LOG(ferr << "My direction is " << WorldCoor(data->dx, data->dy)
              << " and my real_a = " << real_a << endl);
        }
      } else {
        real_a = random->rand_real8() * PI2;
      }
      a = (int) (real_a * ANGLE_DIV / PI2 + 0.5); // Rounding
    }

    /*
    if (show_msg) {
      LOG( ferr << " BEFORE reala=" << real_a << " a=" << a << endl );
      LOG( ferr << " elevation=" << (real_a / PI2 * 360) << endl );
      LOG (collisionEngine.Debug(ferr) );
    }
    */
    collisionEngine.Rotate( a?(24-a):0 );
    /*
    if (show_msg) {
      LOG( ferr << " rotated: " << endl );
      LOG (collisionEngine.Debug(ferr) );
      LOG( ferr << " Add fwd " << endl );
    }
    */

    if(!collisionEngine.Isstuck()) {
      if (data->hasRoute)
        collisionEngine.AddForward();
      else
        collisionEngine.AddForward();
    }

    collisionEngine.Find(a, s);
    /*
    if (show_msg) {
      LOG( collisionEngine.Debug(ferr) );
      LOG( ferr << " a=" << a << " s=" << s << endl );
    }
    */

    // Convert the integral angle to a real direction
    WorldCoor new_coor;
    WorldCoor half_coor;
    {
      real_a += (a * DANGLE);
      if (show_msg) {
        while(real_a > PI2) real_a -= PI2;
        LOG( ferr << " AFTER reala=" << real_a << " a=" << a << endl );
        LOG( ferr << " elevation=" << (real_a / PI2 * 360) << endl );
      }
      int new_x = tank_world_coor.x + (int) (s * cos(real_a) + 0.5);
      int new_y = tank_world_coor.y + (int) (s * sin(real_a) + 0.5);
      if (new_x < 0) new_x = 0;
      if (new_y < 0) new_y = 0;
      if (new_x >= 1024) new_x = 1023;
      if (new_y >= 1024) new_y = 1023;
      new_coor.x = new_x;
      new_coor.y = new_y;
      /*
         for(int i = 1; i <= s; i++) {
         WorldCoor s_coor;
         s_coor.x = tank_world_coor.x + (int) (i * cos(real_a) + 0.5);
         s_coor.y = tank_world_coor.y + (int) (i * sin(real_a) + 0.5);
         AddPredictedData(s_coor, data);
         }
       */
      half_coor.x = tank_world_coor.x + (int)(s * cos(real_a) / 2 + 0.5);
      half_coor.y = tank_world_coor.y + (int)(s * sin(real_a) / 2 + 0.5);
    }
    /*
    if (show_msg) {
      while(real_a > PI2) real_a -= PI2;
      LOG(ferr << "Move tank to " << new_coor << " for a distance of (" << a << ',' << s << ')' << endl);
    }
    */
    tank->SetWayPoint(new_coor);
    tank->SetSpeed( (s + 1) / 2);
    AddPredictedData(half_coor, data);
    //AddPredictedData(new_coor, data);
  }

  LOG(ferr << "End FFEngine::Run()" << endl);
}
