#include "FFEngine.H"
#include "GlobalDataBase.H"
#include "PlayerMarine.H"
#include "EnemyMarine.H"
#include "Clustering.H"
#include "FFData.H"
#include "Random.H"
#include <algorithm>
#include <cmath>

using std::max;
using std::min;

namespace {

  inline WorldCoor GetWayPoint(const WorldCoor &curr, const Force& force, const int &speed) {
    real_type multiplier = (speed + 0.5);
    if (force.real() > EPS)
      multiplier = min(multiplier,
          (WORLD_WIDTH - 1 - curr.x) / force.real() );
    else if (force.real() < -EPS)
      multiplier = min(multiplier, -curr.x / force.real());

    if (force.imag() > EPS)
      multiplier = min(multiplier,
          (WORLD_HEIGHT - 1 - curr.y) / force.imag() );
    else if (force.imag() < -EPS)
      multiplier = min(multiplier, -curr.y / force.imag());

    sint4 new_x = curr.x + (sint4) (multiplier * force.real());
    sint4 new_y = curr.y + (sint4) (multiplier * force.imag());
    return WorldCoor(new_x, new_y);
  }

  typedef long long ll_t;

}

FFEngine::FFEngine(GlobalDataBase* global_data_base):globalDataBase(global_data_base) {
  memset(mapBucketSize, 0, sizeof(mapBucketSize));
  PlayerMarine** player_begin = globalDataBase->GetPlayerMarines();
  PlayerMarine** player_end = globalDataBase->GetPlayerMarinesEnd();

  NeutralObj** neutral_begin = globalDataBase->GetNeutralObjs();
  NeutralObj** neutral_end = globalDataBase->GetNeutralObjsEnd();

  // Install the ffBase into them
  for (PlayerMarine** i = player_begin; i != player_end; ++i)
    (*i)->ffData = new FFPlayerData(this, *i);

  for (NeutralObj** i = neutral_begin; i != neutral_end; ++i)
    (*i)->ffData = new FFNeutralData(this, *i);
}

void FFEngine::Remove(const MapCoor& coor, FFData* data) {
  FFData** list = &mapBucket[coor.r][coor.c][0];
  char &n = mapBucketSize[coor.r][coor.c];
  // I am taking out the loop because it is slow....
  if (list[0] == data) list[0] = list[--n];
  else if (list[1] == data) list[1] = list[--n];
  else if (list[2] == data) list[2] = list[--n];
  else if (list[3] == data) list[3] = list[--n];
  else if (list[4] == data) list[4] = list[--n];
  else --n;
}

void FFEngine::Add(const MapCoor& coor, FFData* data) {
  FFData** list = &mapBucket[coor.r][coor.c][0];
  char &n = mapBucketSize[coor.r][coor.c];
  list[n++] = data;
}

void FFEngine::Run() {
  Clustering* clustering = globalDataBase->GetClustering();

  const Cluster* const* cluster_begin = clustering->GetClusters();
  const Cluster* const* cluster_end = clustering->GetClustersEnd();

  PlayerMarine** player_begin = globalDataBase->GetPlayerMarines();
  PlayerMarine** player_end = globalDataBase->GetPlayerMarinesEnd();

  NeutralObj** neutral_begin = globalDataBase->GetNeutralObjs();
  NeutralObj** neutral_end = globalDataBase->GetNeutralObjsEnd();

  for (PlayerMarine** i = player_begin; i != player_end; ++i)
    (*i)->ffData->Update();

  for (NeutralObj** i = neutral_begin; i != neutral_end; ++i)
    (*i)->ffData->Update();

  Random* random = globalDataBase->GetRandom();
  int counter = 0;

  // So now, calculate all the forces
  for (PlayerMarine** it = player_begin; it != player_end; ++it) {
    ++counter;
    PlayerMarine* marine = *it;

    FFPlayerData* data = (FFPlayerData*) marine->ffData;
    WorldCoor dest = marine->GetDest();
    WorldCoor curr_coor = marine->GetWorldCoor();
    Force force(dest.x - curr_coor.x, dest.y - curr_coor.y);
    Force player_force(ZERO), neutral_force(ZERO);
    if (dest.x != curr_coor.x || dest.y != curr_coor.y)
      force = force * DEST_STRENGTH / abs(force);
    // Rotate a bit
    real_type rnd_angle = (random->rand_real8() - 0.5) * RANDOM_ANGLE;
    force *= Force( cos(rnd_angle), sin(rnd_angle) );

    // Find out the map coor
    MapCoor map_coor = data->mapCoor;
    sint4 r1 = map_coor.r - 2;
    sint4 r2 = map_coor.r + 2;
    sint4 c1 = map_coor.c - 2;
    sint4 c2 = map_coor.c + 2;
    if (r1 < 0) {
      if (curr_coor.y <= WALL_DISTANCE)
        force += WALL_STRENGTH * S_FORCE;
      r1 = 0;
    }
    if (r2 >= MAP_HEIGHT) {
      if (curr_coor.y + WALL_DISTANCE >= WORLD_HEIGHT - 1)
        force += WALL_STRENGTH * N_FORCE;
      r2 = MAP_HEIGHT - 1;
    }
    if (c1 < 0) {
      if (curr_coor.x <= WALL_DISTANCE)
        force += WALL_STRENGTH * E_FORCE;
      c1 = 0;
    }
    if (c2 >= MAP_WIDTH) {
      if (curr_coor.x + WALL_DISTANCE >= WORLD_WIDTH - 1)
        force += WALL_STRENGTH * W_FORCE;
      c2 = MAP_WIDTH - 1;
    }
    
    // Force done... now find the influence from neighbours
    for (int r = r1; r <= r2; ++r)
      for (int c = c1; c <= c2; ++c) {
        FFData** list = &mapBucket[r][c][0];
        for (int k = mapBucketSize[r][c]; k--; )
          if (list[k] != data)
            list[k]->AddForce(data, player_force, neutral_force);
      }

    real_type player_mag = abs(player_force),
              neutral_mag = abs(neutral_force);
    if (player_mag > EPS)
      force += PLAYER_STRENGTH * player_force / player_mag;
    if (neutral_mag > EPS)
      force += NEUTRAL_STRENGTH * neutral_force / neutral_mag;

    // Force from neighbouring objects done
    // Now, calculate the repulsive force from the hulls
    for (const Cluster* const* cluster = cluster_begin;
        cluster != cluster_end; ++cluster)
      force += FFEngine::GetHullForce( *cluster, marine, data );

    // I have the force, move!
    // See how strong the force is
    real_type force_mag = abs(force);
    if (force_mag >= SPEED_1) force_mag /= force_mag;

    if (force_mag < SPEED_2) {
      if (force_mag < SPEED_1) {
        // Don't move
        marine->SetSpeed(0);
        marine->SetWayPoint(marine->GetWorldCoor());
      } else {
        // Move at speed 1
        marine->SetSpeed(1);
        marine->SetWayPoint(GetWayPoint(marine->GetWorldCoor(), force, 1));
      }
    } else {
      if (force_mag < SPEED_3) {
        // Move at speed 2
        marine->SetSpeed(2);
        marine->SetWayPoint(GetWayPoint(marine->GetWorldCoor(), force, 2));
      } else {
        // Move at full speed
        marine->SetSpeed(3);
        marine->SetWayPoint(GetWayPoint(marine->GetWorldCoor(), force, 3));
      }
    }

  }
}

Force FFEngine::GetHullForce(const Cluster* const &cluster,
    const PlayerMarine* const &marine, const FFPlayerData* const &data) {

  const WorldCoor *hull_begin = cluster->GetHull();
  const WorldCoor *hull_end = cluster->GetHullEnd();

  // Special cases
  // This should not happen
  // if (hull_begin == hull_end) return ZERO;

  if (hull_begin + 1 == hull_end) {
    return ZERO;
    /*
    // One vertex...., just like a sheep
    int dx = data->worldCoor.x - hull_begin->x;
    int dy = data->worldCoor.y - hull_begin->y;
    int dist = dx * dx + dy * dy;
    if (dist > SQR_CORNER_HULL_DISTANCE) return ZERO;
    return HULL_STRENGTH * Force(dx, dy) / abs(dist);
    */
  }

  if (hull_begin + 2 == hull_end) {
    return ZERO; // Just go all out?
    // Two vertices?
    // basically a rectangular block?
  }

  // Figure out the closest hull feature
  // Is a vertex the closets?
  WorldCoor marine_coor = marine->GetWorldCoor();
  WorldCoor dest = marine->GetDest();

  int closest_vertex_idx = 0;

  int min_vertex_dist = hull_begin->sqr_dist(marine_coor);
  WorldCoor prev, curr, next;
  int dist, cnt = 1;

  // Find the vertex that is closest...
  for (const WorldCoor* i = hull_begin+1; i != hull_end; ++i, ++cnt) {
    dist = i->sqr_dist(marine_coor);
    if (dist < min_vertex_dist) {
      min_vertex_dist = dist;
      closest_vertex_idx = cnt;
    }
  }

  // The closest edge must either be the next or the previous edge
  // Find out the next one...
  if (hull_begin + closest_vertex_idx + 1 != hull_end)
    next = hull_begin[closest_vertex_idx+1];
  else
    next = *hull_begin;
  // Find the previous one
  if (closest_vertex_idx)
    prev = hull_begin[closest_vertex_idx-1];
  else
    prev = hull_end[-1];
  curr = hull_begin[closest_vertex_idx];

  // Figure out where my point is
  int dx = marine_coor.x - curr.x, dy = marine_coor.y - curr.y,
      next_dx = next.x - curr.x, next_dy = next.y - curr.y,
      prev_dx = prev.x - curr.x, prev_dy = prev.y - curr.y;
  int prev_nx = -prev_dy, prev_ny = prev_dx,
      next_nx = next_dy, next_ny = -next_dx;
  Force prev_force(prev_nx, prev_ny), next_force(next_nx, next_ny);
  prev_force /= abs(prev_force); next_force /= abs(next_force);
  real_type real_dist;
  
  if ( dx * prev_nx + dy * prev_ny < 0
      && dx * next_nx + dy * next_ny < 0) {
    // Get away from the center of mass!
    WorldCoor center = cluster->GetCenterOfMass();
    Force ret(marine_coor.x - center.x, marine_coor.y - center.y);
    ret /= abs(ret);
    return IN_HULL_STRENGTH * ret;

    // Within... check the angular bisector
    Force prev_line(prev_dx, prev_dy);
    Force next_line(next_dx, next_dy);
    prev_line /= abs(prev_line);
    next_line /= abs(next_line);
    Force middle = prev_line + next_line;

    if (dx * middle.imag() - dy * middle.real() >= 0)
      // Go the next one
      return IN_HULL_STRENGTH * next_force;
    else
      // Go the prev one
      return IN_HULL_STRENGTH * prev_force;

  } else {
    // Check perpendicular ones
    // See which region it lies
    if (dx * prev_ny - dy * prev_nx >= 0) {
      // Compare this with
      real_dist = dx * prev_nx + dy * prev_ny;
      real_dist /= sqrt(prev_nx * prev_nx + prev_ny * prev_ny);
      if (real_dist > EDGE_HULL_DISTANCE) return ZERO;
      Force hull_force = HULL_STRENGTH * prev_force;
      Force curl_force(prev_dx, prev_dy);
      // Try to curl as well
      int dest_dx = dest.x - marine_coor.x,
          dest_dy = dest.y - marine_coor.y,
          from_prev_dx = marine_coor.x - prev.x,
          from_prev_dy = marine_coor.y - prev.y;
      if ( from_prev_dx * dest_dx + from_prev_dy * dest_dy
          >  dx * dest_dx + dy * dest_dy )
        curl_force = -curl_force;
      curl_force = CURL_STRENGTH * curl_force / abs(curl_force);
      return hull_force + curl_force;

    } else if (next_nx * dy - next_ny * dx >= 0) {
      real_dist = dx * next_nx + dy * next_ny;
      real_dist /= sqrt(next_nx * next_nx + next_ny * next_ny);
      if (real_dist > EDGE_HULL_DISTANCE) return ZERO;
      Force hull_force = HULL_STRENGTH * next_force;
      Force curl_force(next_dx, next_dy);
      // Try to curl as well
      int dest_dx = dest.x - marine_coor.x,
          dest_dy = dest.y - marine_coor.y,
          from_next_dx = marine_coor.x - next.x,
          from_next_dy = marine_coor.y - next.y;
      if ( from_next_dx * dest_dx + from_next_dy * dest_dy
          >  dx * dest_dx + dy * dest_dy )
        curl_force = -curl_force;
      curl_force = CURL_STRENGTH * curl_force / abs(curl_force);
      return hull_force + curl_force;

    } else { // Closest to the vertex...
      if (min_vertex_dist > SQR_EDGE_HULL_DISTANCE) return ZERO;
      
      Force ret( marine_coor.x - curr.x, marine_coor.y - curr.y);
      ret /= abs(ret);
      if ( marine->GetCooldown() > 1 ) { // Push it out
        return HULL_STRENGTH * ret;
      }

      // So the marine is not on cool down, he can engage
      if (min_vertex_dist > SQR_CORNER_HULL_DISTANCE) return ZERO;
      return CORNER_STRENGTH * ret / abs(ret);
    }
  }

  return ZERO; // Shouldn't touch this one though

}
