// $Id: Game4EventHandler.C,v 1.13 2007/03/15 10:37:47 orts_furtak Exp $

// This is an ORTS file (c) Michael Buro, licensed under the GPL

/* 
   Copyright 2007 Man Hon Chan, Zhangbo Liu, Simon Suyadi
   This code is Licensed under GPL.
 */
// A sample client for ORTS that shows how to access client data

// todo/questions:
//
// game scripts:
// - what to do with workers during build time
// - similarly, how to define mining?
// - how is attack range defined (distance between centers/hulls)?
//
// client:
// - how to access global player data?
// - how to compute the set of visible tiles?
// - how to compute tiles visible from a unit?
// - how to decide what can be attacked by a specific unit?

// FORS/FORALL/ERR macros are defined in orts3/libs/kernel/src/Global.H

#include "Game4EventHandler.H"
#include "GameStateModule.H"
#include "Game.H"
#include "GameChanges.H"
#include "GameTile.H"
#include "ServerObjData.H"
#include "Options.H"
#include <algorithm>
#include <cstdio>
#include <iostream>
#include "FFEngine.H"

#ifdef GRAPHICS
#include "GfxModule.H"
#include "GUI.H"
#include "DrawOnTerrain2D.H"
#endif

using namespace std;

namespace {

  static const int NUMBER_OF_SQUADS = 10;
  static const int NUMBER_OF_MARINES = 50;
  const int CLUSTERING_UPDATE_INTERVAL = 2;
  const int GENERAL_AI_UPDATE_INTERVAL = 2;

#ifdef GRAPHICS
  Vec3<real4> WHITE(1.0, 1.0, 1.0);
  Vec3<real4> ORANGE(1.0, 0.5, 0.0);
  Vec3<real4> VIOLET(1.0, 0.0, 1.0);
  Vec3<real4> BLUISH(0.0, 0.3, 1.0);
#endif
}


bool Game4EventHandler::handle_event(const Event& e)
{

  // Post initialize
  if (!postInitialized) {
    PostInitialize();
  }

  if (e.get_who() == GameStateModule::FROM) {

    //cout << "Event: " << e.who() << " " << e.what() << endl;

    if (e.get_what() == GameStateModule::STOP_MSG) {
      state.quit = true; return true;
    } else if (e.get_what() == GameStateModule::READ_ERROR_MSG) {
      state.quit = state.error = true; return true;
    } else if (e.get_what() == GameStateModule::VIEW_MSG) {

      //cerr << "Before Updating" << endl;
      globalDataBase->Update(state.gsm->get_changes());

      sint4 game_tick = globalDataBase->GetGameTick();

      //if ( game_tick - clusteringTime >= CLUSTERING_UPDATE_INTERVAL ) {
      globalDataBase->GetClustering()->Update();
      //clusteringTime = game_tick;
      //}

      if ( game_tick - generalAITime >= GENERAL_AI_UPDATE_INTERVAL) {
        generalAI->Update();
        generalAITime = game_tick;
      }

      if (game_tick >= 80)
        generalAI->FinishAssemble();


      //cerr << "Before computing actions" << endl;
      ComputeActions();
      //cerr << "Before FF" << endl;

      // Do the Force Field
      ffEngine->Run();
      //cerr << "After FF" << endl;

      //cerr << "Before Act" << endl;
      // Set all actions
      PlayerMarine** marine = globalDataBase->GetPlayerMarines();
      PlayerMarine** marine_end = globalDataBase->GetPlayerMarinesEnd();
      for(; marine != marine_end; ++marine) (*marine)->Act();
      // All actions set, now send them
      state.gsm->send_actions();
      //cerr << "Actions sent" << endl;

      // REMARK: Here belows are debugging information

      Game &game = state.gsm->get_game();	

      sint4 vf = game.get_view_frame();
      sint4 af = game.get_action_frame();
      sint4 sa = game.get_skipped_actions();
      sint4 ma = game.get_merged_actions();

#ifdef FLOODING
      if ((vf % 100) == 0)
        globalDataBase->Display();
#endif

      //if ((vf % 20) == 0) cout << "[frame " << vf << "]" << endl;

      if (vf != af || sa || ma > 1) {

        cout << "frame " << vf;
        if (af < 0) cout << " [no action]";
        else if (vf != af) cout << " [behind by " << (vf - af) << " frame(s)]";
        if (sa) cout << " [skipped " << sa << "]";
        if (ma > 1) cout << " [merged " << ma << "]";
        cout << endl;
      }

#ifdef GRAPHICS
      if (state.gfxm) {

        // don't draw if we are far behind
        if (abs(vf-af) < 10) {
          state.gfxm->draw();
          state.just_drew = true;
        }
      }

      if (state.gui) {
        state.gui->event();
        //        state.gui->display();
        if (state.gui->quit) exit(0);
      }

      if (state.gui) {
        // as a hack for flickering, gui->display() is called in dot.start()
        // so call everything that draws to the debug window between start() and end()
        DrawOnTerrain2D dot(state.gui);
        dot.start();
        // Draw all clusters?
        Clustering* clustering = globalDataBase->GetClustering();
        Cluster* const* cluster_begin = clustering->GetClusters();
        Cluster* const* cluster_end = clustering->GetClustersEnd();
        for (Cluster* const* i = cluster_begin; i != cluster_end; ++i) {
          Cluster* const cluster = *i;
          // Get the hulls...
          WorldCoor prev_coor = cluster->GetHullEnd()[-1];
          for (const WorldCoor* j = cluster->GetHull(), *ej = cluster->GetHullEnd();
              j != ej; ++j) {
            const WorldCoor &now = *j;
            dot.draw_line(prev_coor.x, prev_coor.y, now.x, now.y, WHITE);
            prev_coor = now;
          }
          WorldCoor center = cluster->GetCenterOfMass();
          dot.draw_circle(center.x, center.y, 5, VIOLET);
          dot.draw_circle(center.x, center.y, 10, VIOLET);
        }
        // Draw all range
        for (const EnemyMarine* const * j = globalDataBase->GetEnemyMarines(),
            *const *ej = globalDataBase->GetEnemyMarinesEnd(); j!=ej; ++j) {
          WorldCoor marine_coor = (*j)->GetWorldCoor();
          dot.draw_circle(marine_coor.x, marine_coor.y, 64+4, BLUISH);
        }
        generalAI->DrawAssignment(dot);
        dot.end();
      }
#endif

      return true;
    }
  }

  return false;
}


void Game4EventHandler::ComputeActions()
{
  generalAI->Act();
  return;

#ifdef GARBAGE
  cout << endl;

  cout << "compute actions" << endl;

  const GameChanges &changes = state.gsm->get_changes();

  bool cinfo;
  Options::get("-cinfo", cinfo);

  if (cinfo) {

    // - what changed in the last simulation tick

    if (!changes.new_tile_indexes.empty())
      cout << "#new tiles = " << changes.new_tile_indexes.size() << endl;

    if (!changes.new_boundaries.empty())
      cout << "#new boundaries = " << changes.new_boundaries.size() << endl;

    // fixme: what about removed tiles and removed boundaries?
    // currently tiles and boundaries are never removed

    if (!changes.new_objs.empty())
      cout << "#new objects = " << changes.new_objs.size() << endl;

    if (!changes.changed_objs.empty())
      cout << "#changed objects = " << changes.changed_objs.size() << endl;

    if (!changes.vanished_objs.empty())
      cout << "#vanished objects = " << changes.vanished_objs.size() << endl;

    if (!changes.dead_objs.empty())
      cout << "#dead objects = " << changes.dead_objs.size() << endl;

  }

  const Game &game = state.gsm->get_game();
  const sint4 cid = game.get_client_player();

  // ***************** terrain ******************

  {
    const Map<GameTile> &map = game.get_map();

    bool minfo;
    Options::get("-minfo", minfo);

    if (minfo) {

      // - map dimensions

      cout << "Map width: = " << map.get_width()  << " tiles" << endl;
      cout << "Map height: " << map.get_height() << " tiles" << endl;  

      // height field scale: HEIGHT_MULT = one tile width up, constant for now
      cout << "height mult: " << GameConst::HEIGHT_MULT << endl; 

      // - new/updated tile indexes

      cout << "# of new/updated tiles " << changes.new_tile_indexes.size() << endl;
      cout << "Index(es): ";
      FORALL (changes.new_tile_indexes, it) cout << *it << " ";
      cout << endl;

      // - new boundaries

      const Vector<ScriptObj::WPtr> &boundaries = changes.new_boundaries;

      FORALL (boundaries, it) {

        const GameObj *g = (*it)->get_GameObj();
        if (g == 0) continue;

        cout << "new boundary (" << *g->sod.x1 << "," << *g->sod.y1 << ")-("
          << *g->sod.x2 << "," << *g->sod.y2 << ")"
          << endl;
      }

      // - all boundaries

      const Game::ObjCont &all_boundaries = game.get_boundaries();
      cout << "total # of boundaries = " << all_boundaries.size() << endl;
    }


    bool tinfo;
    Options::get("-tinfo", tinfo);

    if (tinfo) {

      // - all tiles

      FORS (y, map.get_height()) {
        FORS (x, map.get_width()) {
          const GameTile &tile = map(x,y);
          cout << "[" << x << "," << y << "]: ";

          sint4 hNW, hNE, hSE, hSW;
          Tile::Split split;

          // get corner heights and split type
          tile.get_topo(hNW, hNE, hSE, hSW, split);

          // tile types: see enum on top
          // split types: see orts3/libs/kernel/src/Tile.H
          // if split == Tile::NO_SPLIT => typeW is the tile type

          cout << "split:" << split 
            << " type w:" << tile.get_typeW()
            << " type e:" << tile.get_typeE()
            << " heights:" << hNW << " " << hNE << " " << hSE << " " << hSW
            << endl;
        }

        cout << endl;
      }

      cout << endl;
    }


    // - currently visible tiles

    // ... to be implemented
  }

  // ***************** objects ******************

  bool oinfo;
  Options::get("-oinfo", oinfo);

  if (oinfo) {

    // units under client player control

    const Game::ObjCont &objs = game.get_objs(cid);

    FORALL (objs, it) {

      const GameObj *gob = (*it)->get_GameObj();
      if (!gob) ERR("not a game object!?");

      const ServerObjData &sod = gob->sod;

      // - fundamental properties

      // blueprint name ~ type name
      cout << "bp: " << gob->bp_name();

      if (!sod.in_game) { cout << " no gameobj" << endl; continue; }

      if (*sod.x < 0) { cout << "not on playfield" << endl; continue; }

      // object on playfield

      // sod.shape and sod.zcat values defined in orts3/libs/kernel/src/Object.H

      cout << "zcat: " << *sod.zcat
        << " max-speed: " << *sod.max_speed
        << " speed: " << *sod.speed 
        << " moving: " << *sod.is_moving 
        << " sight: " << *sod.sight;

      if (*sod.shape == Object::CIRCLE) {
        cout << " CIRCLE loc: (" << *sod.x << "," << *sod.y << ")"
          << " r: "<< *sod.radius;

      } else if (*sod.shape == Object::RECTANGLE) {
        cout << " RECT dim: (" << *sod.x1 << "," << *sod.y1 << ")-("
          << *sod.x2 << "," << *sod.y2 << ")";
      } else {
        cout << " UNKNOWN SHAPE";
      }

      // - other object properties

      // heading: 0 = stopped, 1...GAME_CONST::HEADING_N
      const sintptr *hd = gob->get_int_ptr("heading");
      if (hd) cout << " hd: " << GameConst::angle_from_dir(*hd, GameConst::HEADING_N);

      // ... more examples

      cout << endl;
    }
  }

  // - what an object currently sees

  // ... to be implemented

  // - setting object actions

  const Map<GameTile> &map = game.get_map();
  const Game::ObjCont &objs = game.get_objs(cid);
  const sint4 points_x = map.get_width()  * game.get_tile_points();
  const sint4 points_y = map.get_height() * game.get_tile_points();  

  FORALL (objs, it) {

    GameObj *gob = (*it)->get_GameObj();

    if (gob && gob->sod.in_game && *gob->sod.x >= 0 && !gob->is_dead()) {

      ScriptObj *weapon = gob->component("weapon");
      if (!weapon) continue;

      const sint4 n = game.get_player_num();
      GameObj *target = 0;

      if (n > 1) { // there are other players

        // todo: aim at weakest opponent unit within range

        sintptr *prange = weapon->get_int_ptr("max_ground_range");
        if (!prange) continue;

        const sint4 range2 = *prange * *prange;

        FORS (i, n) {

          if (i == cid) continue;

          // check for targets (could be done much faster)

          const Game::ObjCont &opp_objs = game.get_objs(i);

          FORALL (opp_objs, it) {

            GameObj *gob2 = (*it)->get_GameObj();
            if (gob2 && gob2->sod.in_game && *gob2->sod.x >= 0 && !gob2->is_dead()) {
              real8 d2 =
                square(real8(*gob->sod.x - *gob2->sod.x)) +
                square(real8(*gob->sod.y - *gob2->sod.y));

              if (d2 <= range2) {
                // target within range
                target = gob2;
              }
            }
          }
        }
      }

      if (target) {

        cout << "attack" << endl;

        // attack target

        // fixme: observe cooldown

        //cout << "launch " << sqrt(min_d) << " " << sqrt(attack_d) << endl;
        Vector<sint4> args;
        args.push_back(game.get_cplayer_info().get_id(target));
        gob->component("weapon")->set_action("attack", args);
        continue;
      }

      if (*gob->sod.speed == 0) {

        // object not dead and speed == 0 => move action

        // todo: attack weakest opponent in range

        ScalarPoint t;

        if ((random.rand_uint4() % 100) < 33) {

          // random target on fringe (x or y unchanged)

          switch (random.rand_uint4() & 3) {
            case 0: t.x = *gob->sod.x; t.y = 0; break;
            case 1: t.x = *gob->sod.x; t.y = points_y-1; break;
            case 2: t.x = 0; t.y = *gob->sod.y; break;           
            case 3: t.x = points_x-1; t.y = *gob->sod.y; break;
          }

        } else {

          // random target on fringe

          switch (random.rand_uint4() & 3) {

            case 0:
              t.x = 0;
              t.y = random.rand_uint4() % points_y;
              break;

            case 1:
              t.x = points_x - 1;
              t.y = random.rand_uint4() % points_y;
              break;

            case 2:
              t.x = random.rand_uint4() % points_x;
              t.y = 0;
              break;

            case 3:
              t.x = random.rand_uint4() % points_x;
              t.y = points_y - 1;
              break;
          }
        }

        Vector<sint4> params;
        params.push_back(t.x);
        params.push_back(t.y);
        params.push_back(*gob->sod.max_speed);

        gob->set_action("move", params);
      }
    }
  }

  // - global player/game variables

  // ... to be implemented
#endif
}

void Game4EventHandler::PostInitialize() {

  Game &game = state.gsm->get_game();

  globalDataBase = new GlobalDataBase(game, &random);
  //ffEngine = new FFEngine(globalDataBase);
  globalDataBase->GetClustering()->Update();
  clusteringTime = 0;
  generalAI = new GeneralAI(globalDataBase);
  generalAI->Init();
  generalAITime = 0;
  ffEngine = new FFEngine(globalDataBase);
  postInitialized = true;
}


