#include "Regrouping.H"
#include <algorithm>
#include <vector>
#include "PlayerMarine.H"
/* 
   Copyright 2007 Man Hon Chan, Zhangbo Liu, Simon Suyadi
   This code is Licensed under GPL.
 */

using std::max;

using std::cerr;
using std::endl;

namespace {
  const int SQR_THRESHOLD = 48*48;
  const int MAX_UNIT_COUNT = 50;
  int parent[MAX_UNIT_COUNT];
  int size[MAX_UNIT_COUNT];
  int renaming[MAX_UNIT_COUNT];

  inline void InitUnionFind(int n) {
    for (int i = 0; i < n; ++i)
      size[parent[i] = i] = 1;
  }

  int Find(int x) {
    return (x == parent[x])?x:(parent[x] = Find(parent[x]));
  }

  inline void Merge(int x, int y) {
    x = Find(x); y = Find(y);
    if (x == y) return; // We are all happy
    if (size[x] >= size[y]) {
      parent[y] = x;
      size[x] += size[y];
      size[y] = 0;
    } else {
      parent[x] = y;
      size[y] += size[x];
      size[x] = 0;
    }
  }

  // Various functions for chopping
  bool OrderX(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    return a_coor.x < b_coor.x;
  }

  bool OrderY(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    return a_coor.y < b_coor.y;
  }

  bool OrderXPY(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    return a_coor.x + a_coor.y < b_coor.x + b_coor.y;
  }

  bool OrderXMY(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    return a_coor.x - a_coor.y < b_coor.x - b_coor.y;
  }

  sint4 DistX(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    const int dx = a_coor.x - b_coor.x, dy = a_coor.y - b_coor.y;
    return 2 * dx * dx;
  }

  sint4 DistY(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    const int dx = a_coor.x - b_coor.x, dy = a_coor.y - b_coor.y;
    return 2 * dy * dy;
  }

  sint4 DistXPY(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    const int dx = a_coor.x - b_coor.x, dy = a_coor.y - b_coor.y;
    return (dx + dy) * (dx + dy);
  }

  sint4 DistXMY(const PlayerMarine* const &a, const PlayerMarine* const &b) {
    const WorldCoor& a_coor = a->GetWorldCoor();
    const WorldCoor& b_coor = b->GetWorldCoor();
    const int dx = a_coor.x - b_coor.x, dy = a_coor.y - b_coor.y;
    return (dx - dy) * (dx - dy);
  }

  typedef bool OrderFunc(const PlayerMarine* const &, const PlayerMarine* const &);
  typedef sint4 DistFunc(const PlayerMarine* const &, const PlayerMarine* const &);

  sint4 SplitScore(std::vector<PlayerMarine*> &marines, OrderFunc *order, DistFunc *dist) {
    // n-th element
    int n = marines.size();
    if (n < 8) {
      cerr << "Why is someone passing in a vector with less than 8 elements?! " << n << endl;
    }
    std::nth_element(marines.begin(), marines.begin() + (n / 2), marines.end(), order);
    const PlayerMarine* const& median = marines[n/2];
    sint4 result = 0;
    for (int i = 0; i < n; ++i) result += dist(marines[i], median);
    return result;

  }

  /*
  bool OrderBySize(const SquadAI* const &a, const SquadAI* const &b) {
    return a->NumOfMarines() > b->NumOfMarines();
  }
  */

}

Regrouping::Regrouping(GlobalDataBase* global_data_base):globalDataBase(global_data_base), squadCount(0) {
  // Create all the squads
  for (int i = 0; i < MAX_GROUP_COUNT; ++i)
    squads[i] = new SquadAI(globalDataBase);
}

void Regrouping::Regroup() {
  //cerr << "Regrouping::Regroup()" << endl;
  int n = globalDataBase->GetPlayerMarineCount();
  InitUnionFind(n);
  int i = 0, j = 0;

  for (PlayerMarine** marine = globalDataBase->GetPlayerMarines(),
      **marine_end = globalDataBase->GetPlayerMarinesEnd();
      marine != marine_end; ++marine, ++i) {
    j = i+1;
    for (PlayerMarine** next = marine+1; next != marine_end; ++next, ++j)
      if ( (*marine)->GetWorldCoor().sqr_dist( (*next)->GetWorldCoor() )
          < SQR_THRESHOLD )
        // Merge them
        Merge(i, j);
  }

  // We can now find the parent group of each enemy unit
  // So we want to rename all group
  squadCount = 0;
  for (int k = 0; k < n; ++k)
    if (size[k]) renaming[k] = squadCount++;
  for (int k = 0; k < squadCount; ++k) squads[k]->Reset();
  
  // First mindlessly assign all the squads
  i = 0;
  for (PlayerMarine** marine = globalDataBase->GetPlayerMarines(),
      **marine_end = globalDataBase->GetPlayerMarinesEnd();
      marine != marine_end; ++marine, ++i) {
    int squad_id = renaming[ Find(i) ];
    squads[squad_id]->AddMarine(*marine);
  }

  for (int k = 0; k < squadCount;) {
    SquadAI* squad = squads[k];
    if ( squad->NumOfMarines() < 8 ) {
      ++k;
      continue;
    }
    // Split it into groups
    SquadAI* squad2 = squads[squadCount++];
    squad2->Reset();

    std::vector<PlayerMarine*> &squadMarines = squad->squadMarines;

    // For all squad members
    j = squadMarines.size();

    // We have four principle split, and we want to see which one is best
    sint4 score_x = SplitScore(squadMarines, OrderX, DistX),
          score_y = SplitScore(squadMarines, OrderY, DistY),
          score_xpy = SplitScore(squadMarines, OrderXPY, DistXPY),
          score_xmy = SplitScore(squadMarines, OrderXMY, DistXMY);
    sint4 best_score = max( max(score_x, score_y), max(score_xpy, score_xmy) );

    j /= 2;
    if (best_score == score_x) {
      std::nth_element(squadMarines.begin(), squadMarines.begin() + j, squadMarines.end(), OrderX);
      while (squadMarines.size() > j) {
        squad2->AddMarine(squadMarines.back());
        squadMarines.pop_back();
      }
    } else if (best_score == score_y) {
      std::nth_element(squadMarines.begin(), squadMarines.begin() + j, squadMarines.end(), OrderY);
      while (squadMarines.size() > j) {
        squad2->AddMarine(squadMarines.back());
        squadMarines.pop_back();
      }
    } else if (best_score == score_xpy) {
      std::nth_element(squadMarines.begin(), squadMarines.begin() + j, squadMarines.end(), OrderXPY);
      while (squadMarines.size() > j) {
        squad2->AddMarine(squadMarines.back());
        squadMarines.pop_back();
      }
    } else { // best score = score_xmy
      std::nth_element(squadMarines.begin(), squadMarines.begin() + j, squadMarines.end(), OrderXMY);
      while (squadMarines.size() > j) {
        squad2->AddMarine(squadMarines.back());
        squadMarines.pop_back();
      }
    }

  }

  // std::sort(squads, squads + squadCount, OrderBySize);
}

void Regrouping::Update() {
  // Clear dead squads
  for (int i = 0; i < squadCount; )
    if (squads[i]->IsAlive())
      ++i;
    else
      std::swap(squads[i], squads[--squadCount]);
  // Sort them
  // std::sort(squads, squads + squadCount, OrderBySize);
}
