/* Schedules and Autonomously runs a tournament. 
 * 
 * Krysta Mirzayans  June 7/06
 */

#include "tmanagerNEW.h"


//GLOBAL GAME DATA STRUCTURES:
map<int, GAME> game_queue;      //list of games to be played

//COMPUTERS:
map<int, COMPUTER> computers;  

//GLOBAL PLAYERS DATA STRUCTURE:
map<int, PLAYERDATA> players;   //indexed starting from 0

//THREAD VARIABLES: 
pthread_t *threads_game;        //indexed using game id          
pthread_mutex_t mutexGames;   

/**
 * Error reporting function, which displays the given error
 * message to standard out and terminates the program with code 20.
 *
 * @param msg  error message to be displayed
 */
void Error(string msg)
{
  if(msg == "usage")
    {
      cerr<< "usage: tmanager -help | configFile\n\
 -help       : print a detailed description \n\n";

      exit(20);
    }
  cerr << "*** " << msg << endl;
  
  exit(20);
}


/**
 * Helper function for parsing the configuration file which
 * skips sections of comments (denoted by #).
 * 
 * @param in  the input file stream (configuration file)
 * @return    a bool flag indicating whether or not end-of-file is reached (true - eof, false - otherwise)
 */
bool skipComments(ifstream &in)
{
  string s;
  char c;
  FOREVER
    {
      if((c = in.peek())== EOF) return true;
      if((c != '#') && (c != ' ') && (c != '\n')) break;   	
      getline(in, s);
    }
  return false;
}


/**
 * Displays the help message to standard out when -help is the command line argument.
 * Exits the program with code 10.
 */
void print_help()
{
  cout << "tmanager - an ORTS tournament manager.                       \n";
  cout << "It reads a list of players and a list of games played so far \n\
and schedules and autonomously runs the tournament.                     \n";
  cout << "\nTournament-file syntax:\n";
  cout << "# comment                                                    \n\
# general section                                                       \n\
<game type (1,2,or3)>                                                   \n\
<expected length of game in minutes>                                    \n\
<allowable grace period for game length in minutes>                     \n\
<starting map seed>                                                     \n\
<number-of-games-per-pair>                                              \n\
s|d # (single: one game per map or double: two games per map)           \n\
\n# player section                                                      \n\
<name> <username>                                                       \n\
...                                                                     \n\
%                                                                       \n\
# computer section                                                      \n\
# generated automatically with -genComps, must be provided otherwise    \n\
<computer1>                                                             \n\
<computer2>                                                             \n\
...                                                                     \n\
%                                                                       \n\
\n# game section                                                        \n\
<map-seed> <name1> <name2> <arbitrary-result-string>                    \n\ 
...                                                                     \n";
  cout << "\nSample tournament-file:\n";
  cout << "\n#header comments                                           \n\
4   # 4 games per pair                                                  \n\
d   # double: two games / map                                           \n\
\nUofA  23411832                                                        \n\
USC   41253533                                                          \n\
UMich 23435223                                                          \n\
\n%                                                                     \n\
# computer section                                                      \n\
ug01                                                                    \n\
ug02                                                                    \n\
ug03                                                                    \n\
ug04                                                                    \n\
%                                                                       \n\
\n# games                                                               \n\
\n# game seed for map i computed as XOR(all player seeds) + i           \n\
# if double, maps are used for two games                                \n\
\n234512345 UofA  USC   0 1 0 3  2 0 2.5 0                        \n\
234512345 USC   UofA  0 1 0 3  2 0 2.5 0                          \n\
234512346 UofA  USC   0 1 0 3  2 0 2.5 0                          \n\
234512346 USC   UofA  0 1 0 3  2 0 2.5 0                          \n\
\n234512345 UofA  UMich 0 1 0 3  2 0 2.5 0                        \n\
234512345 UMich UofA  0 1 0 3  2 0 2.5 0                          \n\
234512346 UofA  UMich 0 1 0 3  2 0 2.5 0                          \n\
234512346 UMich UofA  0 1 0 3  2 0 2.5 0                          \n\
\n234512345 USC   UMich 0 1 0 3  2 0 2.5 0                        \n\
234512345 UMich USC   0 1 0 3  2 0 2.5 0                          \n\
234512346 USC   UMich 0 1 0 3  2 0 2.5 0                          \n\
234512346 UMich USC   0 1 0 3  2 0 2.5 0                          \n\n"; 
  exit(10);
}


/**
 * Displays all players' data to standard out, as read in from the configuration file.
 *
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param playerCount     the total number of players 
 */
void displayPlayers(bool single, int games_per_pair, int playerCount)
{
  int i;
  cout << "\nAll info obtained about players from config file:\n\n";
  cout << "single? " << single << "\tgames_per_pair: " << games_per_pair << endl << endl;
  FOR(i, playerCount)
    {
      cout << "Name:\t\t" << players[i].name << endl;
      cout << "Username:\t" << players[i].username << endl << endl;
    }
}


/**
 * Displays all games in the queue. Shows game index, game seed, player1 name, player2 name
 * @param gameCount  the number of games in the queue
 * @param gameType   the type of game(1,2,or3)
 */
void displayGames(int gameCount, int gameType)
{
  int i;
  cout << "ALL GAMES:" << endl;
  cout << "ID:\t" << "SEED:\t\t" << "PLAYER1:\t";
  if(gameType != 1)
    cout << "PLAYER2:";
  cout << endl;
  FOR(i, gameCount)
    {
      cout << game_queue[i].id << "\t" << game_queue[i].seed << "\t";
      cout << game_queue[i].player_list[0].name;
      if(gameType != 1)
	cout << "\t\t" << game_queue[i].player_list[1].name;
      cout << endl;
      
    }
}


/**
 * Generates the game queue.
 * 
 * @param playerCount     the total number of players
 * @param games_per_pair  the number of games to be played per pair of players (in a given round)
 * @param single          flag indicating whether one or two games are played per map (true-one game, false-two games)
 * @param game_seed       base map seed found by XORing all players' seeds
 * @param gameType        the type of game (1, 2, or 3)
 * @return                total number of games in game_queue
 */
int buildGameList(int games_per_pair, bool single, int playerCount, int game_seed, int gameType)
{
  int gameCount = 0;
  int i,j,k;
  if(gameType == 1)
    {
      FOR(i, playerCount)
	FOR(j, games_per_pair)
	{
	  game_queue[gameCount];
	  game_queue[gameCount].id = gameCount;  
	  game_queue[gameCount].gameType = gameType;
	  game_queue[gameCount].progress = queued; 
	  game_queue[gameCount].saved = false;
	  game_queue[gameCount].seed = game_seed;
	  game_queue[gameCount].player_list[0] = players[i];
	  ++game_seed;
	  ++gameCount;
	}
    }
  else
    {
      FOR(i, playerCount)
	for(j=1; j < playerCount-i; ++j)
	  {
	    FOR(k, games_per_pair)
	      {
		game_queue[gameCount];
		game_queue[gameCount].id = gameCount; 
		game_queue[gameCount].gameType = gameType;
		game_queue[gameCount].progress = queued; 
		game_queue[gameCount].saved = false;
		game_queue[gameCount].seed = game_seed;
	    
		if(!single && k&1)
		  {
		    game_queue[gameCount].player_list[0] = players[j+i];
		    game_queue[gameCount].player_list[1] = players[i];
		  }
		else
		  {
		    game_queue[gameCount].player_list[0] = players[i];
		    game_queue[gameCount].player_list[1] = players[j+i];
		  }

		if(single || k&1)  
		  ++game_seed;
		++gameCount;
	      }
	  }
    }
  
  return gameCount;
}


/**
 * Makes the ssh call to a client and stores the result of this call 
 * as this client's result for this particular game. (not fully implemented yet)
 * Writes all received results to disk files.
 *
 * @param arg  pointer to the game that is being played
 */
void *playClient(void *arg)
{ 
  pthread_mutex_lock(&mutexGames);
  GAME_ARG *gameArg = static_cast<GAME_ARG*>(arg);
  GAME *game = gameArg->game;
  int playerIndex = gameArg->playerIndex;

  //get game information:
  int seed = game->seed;
  PLAYERDATA player = game->player_list[playerIndex];
  COMPUTER *client = game->clients[playerIndex];
  COMPUTER *server = game->server;  //clients need information about which server to connect to
  int id = game->id;
  int gameType = game->gameType;
  timeval start_time = game->start_time;
  pthread_mutex_unlock(&mutexGames);

  char c;
  string s = "ssh -x " + player.username + "@";  //instead of krysta should be player.name!
  s += client->name + " ./WORK/PLAYER"; // + "-server " + server->name + "-id " + player.name + " 2>&1";
s+= " 2>&1";
  FILE *fpREAD =popen(s.c_str(),"r");
  
  s = "";
  while(1)
    {
      c = fgetc(fpREAD);
      if(c==EOF) break;
      s += c;
    }
  pclose(fpREAD);
    
  //lock game_queue to update game_queue info:
  pthread_mutex_lock(&mutexGames);
  gettimeofday(&(game->end_times[playerIndex+1]),NULL);
  timeval end = game->end_times[playerIndex+1];
  game->result[playerIndex+1] = s;
  pthread_mutex_unlock(&mutexGames);

  ostringstream oss;
  oss << "./playerResults/";
  if(!(oss << seed))
    Error("can't make player file");
  oss << "_" << game->player_list[0].name;
  if(gameType != 1)
    oss << "_" << game->player_list[1].name;
 
  ofstream out((oss.str()).c_str(), ios::app);
  if(!out.is_open()) Error("unable to open the disk file for writing");
  out << "__________________________________________________\n";
  out << "Player: " << player.name << "\n(username: " << player.username << ")\n";
  out << "(computer: " << client->name << ")\n";
  tm *timeinfo = localtime(&start_time.tv_sec);
  out << "start time:\t" << asctime(timeinfo);
  timeinfo = localtime(&end.tv_sec);
  out << "end time:\t" << asctime(timeinfo) << endl; 
  out << "Result:\n" << s << endl;
  out.close();

  pthread_exit(NULL);
}


/**
 * Makes the ssh call to the server and stores the result of this call 
 * as the server's result for this particular game. (not fully implemented yet)
 * Writes all received results to disk files.
 * 
 * @param arg  pointer to the game that is being played
 */
void *playServer(void *arg)
{   
  pthread_mutex_lock(&mutexGames);
  GAME *game = static_cast<GAME*>(arg);
  int gameIndex = game->id;
  int seed = game->seed;
  int gameType = game->gameType;
  COMPUTER *server = game->server;
  PLAYERDATA player1 = game->player_list[0];
  if(gameType != 1)
    PLAYERDATA player2 = game->player_list[1];
  pthread_mutex_unlock(&mutexGames);

  char c;
  string s = "ssh -x krysta@";         // what is the server's username?????
  s += server->name + " ./WORK/SERVER";// + "-ids " + player1.name;
  //  if(gameType != 1)
  //  s += " " + player2.name;
  s += " 2>&1";
  
  FILE *fpREAD =popen(s.c_str(),"r");
  
  s = "";
  while(1)
    {
      c = fgetc(fpREAD);
      if(c==EOF) break;
      s += c;
    }
  pclose(fpREAD);

  //lock game_queue to update game_queue info:
  pthread_mutex_lock(&mutexGames);
  gettimeofday(&(game->end_times[0]),NULL);
  game->result[0] = s;
  pthread_mutex_unlock(&mutexGames);
 
  pthread_exit(NULL);
}


/**
 * Sets up the server and clients and spawns the corresponding threads.
 * Updates progress flags for the game & taskCount on the computers 
 *
 * @param arg         pointer to the game that is to be played 
 */
void *playGame(void *arg)
{
  GAME *game = static_cast<GAME*>(arg);
  int rc, status;

  //get game type & update game progress flag
  pthread_mutex_lock(&mutexGames); 
  int gameType = game->gameType;
  game->progress = pending;
  gettimeofday(&(game->start_time),NULL);
  pthread_mutex_unlock(&mutexGames);

  //create the server and client threads

  rc = pthread_create(&(game->thread_srv), NULL, playServer, static_cast<void*>(game));
  if(rc)  Error("cannot create another thread");

  GAME_ARG gameArg1;
  gameArg1.game = game;
  gameArg1.playerIndex = 0;
  rc = pthread_create(&(game->threads_ply[0]), NULL, playClient,static_cast<void*>(&gameArg1));
  if(rc)  Error("cannot create another thread");

  if(gameType != 1)
    {
      GAME_ARG gameArg2;
      gameArg2.game = game;
      gameArg2.playerIndex = 1;
      rc = pthread_create(&(game->threads_ply[1]), NULL, playClient,static_cast<void*>(&gameArg2));
      if(rc)  Error("cannot create another thread");
    }

  //join the server and client threads
  pthread_join( game->thread_srv, (void**)(&status));
  pthread_join( game->threads_ply[0], (void**)(&status));
  if(gameType != 1)
    pthread_join( game->threads_ply[1], (void**)(&status));

  //check for error
  //  if(erno != 0 )
  //  {
      //send signals to individual threads, output error info in file???
  //  }
  //else
  //  {
      //lock game_queue to update game_queue info:
      pthread_mutex_lock(&mutexGames); 
      gettimeofday(&game->end_times[0],NULL);
      game->progress = done;
      game->server->taskCount--;
      if(game->server->taskCount <= 0)
	game->server->isServer = false;
      game->clients[0]->taskCount--;
      if(gameType != 1)
	game->clients[1]->taskCount--;
      pthread_mutex_unlock(&mutexGames);

      ostringstream oss;
      oss << "./playerResults/";
      if(!(oss << game->seed))
	Error("can't make player file");
      oss << "_" << game->player_list[0].name;
      if(gameType != 1)
	oss << "_" << game->player_list[1].name;
 
      ofstream out((oss.str()).c_str(), ios::app);
      if(!out.is_open()) Error("unable to open the disk file for writing");
      out << "__________________________________________________\n";
      out << "Server: " << game->server->name << endl;
      tm *timeinfo = localtime(&game->start_time.tv_sec);
      out << "start time:\t" << asctime(timeinfo);
      timeinfo = localtime(&game->end_times[0].tv_sec);
      out << "end time:\t" << asctime(timeinfo) << endl; 
      out << "Result:\n" << game->result[0] << endl;
      out.close();
      //    }
  pthread_exit(NULL);

}


/**
 * Updates the contents of the configuration file by appending new game results to the existing file.
 *
 * @param filename        the name of the configuration file
 * @param playerCount     the total number of players
 * @param single          flag indicating one or two games per map
 * @param gameCount       the total number of games in the queue
 * @param gameType        the type of game (1, 2, or 3)
 */
void updateConfig(string filename, bool single, int playerCount, int gameCount, int gameType)
{
  int i;
  ofstream out(filename.c_str(), ios::app); // <---
  if(!out.is_open()) Error("unable to open the tournament file for writing");
  
  FOR(i, gameCount)
    {
      if((game_queue[i].progress == done) && !(game_queue[i].saved))
	{ 
	  out << "[" << i << "]  " << game_queue[i].seed << "\n";
	  out << game_queue[i].player_list[0].name << " " << game_queue[i].clients[0]->name << "\t";
	  if(gameType != 1)
	    out << game_queue[i].player_list[1].name << " " << game_queue[i].clients[1]->name << "\t";
	  out << game_queue[i].server->name << "\n" << game_queue[i].result[0] << endl;
	  	  
	  game_queue[i].saved = true;
	
	}  
    }
  out.close();
}


//maybe this should be the first thing called in getNextQueuedGame -- could return if all are queued or not??? saves next one some work
//do you have to lock the matrix in this function ?????!!!!
/**
 * Checks pending games for completed ones.
 *
 * @param gameCount   the total number of games in game_queue 
 * @param game_length the expected length of a game
 * @param epsilon     the allowable grace period for a game
 * @return            indicates whether or not all games are done being played
 */
bool checkGames(int gameCount, double game_length, double epsilon)
{
  int i;
  bool doneAll = true;
  timeval tim;
  timeval start;
  double end_time;
  double start_time;
  FOR(i, gameCount)
    {
      if(game_queue[i].progress == pending)
	{
	  doneAll = false;	
 
	  //check if this game has gone over the time limit:
	  gettimeofday(&tim, NULL);
	  end_time = tim.tv_sec+(tim.tv_usec/1000000.0);
	  start = game_queue[i].start_time;
	  start_time = start.tv_sec+(tim.tv_usec/1000000.0);

	  if((end_time - start_time)/60.0 > game_length+epsilon)
	    {
	      //taken too long -- kill the game & mark as queued
	      game_queue[i].progress = queued;
	      
	    }
	}
    }
  return doneAll;
}

/**
 * Returns the next queued game.
 *
 * @param  gameCount      the total number of games in game_queue
 * @return                the index of the next game in game_queue that is queued (-1 <=> no queued game)
 */
int getNextQueuedGame(int gameCount)
{
  int i;
  FOR(i, gameCount)
    { 
      if(game_queue[i].progress == queued)
	  return i;
    }
  return -1;
}


/**
 * Checks for available computers.
 * If there are available ones, assign them to new games and mark as unavailable
 * 
 * @param compCount   the total number of computers that were initially available 
 * @param gameCount   the total number of games in game_queue
 * @return            Return true if we are done all the games.
 */ 
bool checkComputers(int compCount, int gameCount)
{
  // For now let's assume we can only do one thing per computer
  string s;
  int up;
  int compsNeeded;
  if(game_queue[0].gameType == 1)
    compsNeeded = 2;
  else
    compsNeeded = 3;
  int i;
  int freeComps = 0;
  int nextGame = getNextQueuedGame(gameCount);
  cout << "next queued game: " << nextGame << endl;
  if(nextGame == -1)
    {
       FOR(i, gameCount)
	 {
	   if(game_queue[i].progress == pending)
	     {cout << "game " << i << " PENDING\n";
	       return false;
	     }
	 }
	   return true;
    }
 

  FOR(i, compCount)
    {   
      if(computers[i].taskCount == 0 && computers[i].isUp)
	{ 
	  cout << "check computers!!\t" << i << endl;
	  //check that this computer is up before assigning it:
	  //only works on a ugrad machine
 	  s = "ruptime | grep " +computers[i].name+ " | grep up | wc -l > compCheck"; 
 	  system(s.c_str()); 
 	  ifstream in("compCheck"); 
 	  if(!in.is_open()) Error("unable to open the compCheck file for reading"); 
 	  if(!(in >> up)) Error("improper format in compCheck file"); 
	  system("rm -f ./compCheck");
 	  if(up < 1) 
 	    { 
	      computers[i].isUp = false; 
 	      continue;  
	    }   

	  switch(freeComps)
	    {
	    case 0:
	      game_queue[nextGame].server = &computers[i];
	      game_queue[nextGame].server->taskCount++;
	      game_queue[nextGame].server->isServer = true;
	      break;
	    case 1:
	      game_queue[nextGame].clients[0] = &computers[i];	
	      game_queue[nextGame].clients[0]->taskCount++;  
	      game_queue[nextGame].clients[0]->isServer = false;
	      break;
	     case 2:
	      game_queue[nextGame].clients[1] = &computers[i];	
	      game_queue[nextGame].clients[1]->taskCount++;
	      game_queue[nextGame].clients[1]->isServer = false;
	      break;
	    }

	  ++freeComps;

	}

      if(freeComps >= compsNeeded)
	{
	  int rc = pthread_create(&threads_game[nextGame], NULL, playGame, static_cast<void*>(&game_queue[nextGame]));
	  if(rc)
	    Error("Cannot create another thread for game");
	  return false;
	}

    }
   
  return false;
} 
/**
 * Main function
 */
int main(int argc, char **argv)
{
  int argi, i, j;
  string s;
  char c;

  int playerCount = 0;              //total number of players
  int compCount = 0;                //total number of computers
  int gameCount = 0;                //total number of games in tournament
  int games_per_pair;               //how many games must be played between each pair of players
  int game_seed;                    //base map seed found by XORing all players' seeds
  double game_length;                  //expected length of a game in minutes       
  double epsilon;                   //allowable grace period for game length
  bool single = true;               //1(if true) or 2(if false) games per map
  int gameType = 0;                 //1 - type1, 2 - type2, 3-type3
  bool opt_genComps = false;        //true=> generate computer list for ug lab; false=>read in comps from config file
                                    // ***MAYBE COMMAND LINE SHOULD BE FOLLOWED BY THE LAB???? UG ???
  if (argc < 2) 
    Error("usage");

  argi = 1;                         //start with the first command line argument
 

  // ***COME BACK: add more command line options, such as displaying options
  FOREVER                     //***COME BACK: add more command line options
    { 

      if(argi >= argc) Error("usage");
      if (!strcmp(argv[argi], "-help"))
	print_help();
      else if(!strcmp(argv[argi], "-genComps"))
	{
	  opt_genComps = true;
	  ++argi;
	  continue;
	}
        break;
    }
  
  //PARSING THE INPUT CONFIGURATION FILE:
  ifstream in(argv[argi]);
  if(!in.is_open()) Error("unable to open the tournament file for reading");
  if(skipComments(in)) Error("end of file?");

  //game type & times & seed:
  if(!(in >> gameType)) Error("improper file format -- game type?");
  if(gameType < 1 || gameType > 3)  Error("invalid game type (1,2,or3)");
  if(skipComments(in)) Error("end of file?");

  if(!(in >> game_length)) Error("improper file format -- game length?");
  if(skipComments(in)) Error("end of file?");

  if(!(in >> epsilon)) Error("improper file format -- game length grace period?");
  if(skipComments(in)) Error("end of file?");

  if(!(in >> game_seed)) Error("improper file format -- game seed");
  if(skipComments(in)) Error("end of file?");

  //games_per_pair:
  if(!(in >> games_per_pair)) Error("improper file format -- games per pair?");
  if(skipComments(in)) Error("end of file?");

  //single or double games per map:
  getline(in,s);
  switch(s[0]) 
    {           //***COME BACK: junk the rest of the line: could do a comment check after the first character!!!!!
    case 's':
      break;
    case 'd':
      single = false;
      break;
    default:
      Error("improper file format -- single or double games per map?");
    }
  
  //players and player information (format: name seed ...):
  if(skipComments(in)) Error("end of file?");

  FOREVER
    {
      if((c = in.peek()) == EOF) Error("end of file?");
      in >> s;
      if(!s.empty())
	{ 
	  if(s[0] == '%') break;
	  if(s[0] != '#')
	    {
	      players[playerCount];
	      players[playerCount].index = playerCount;
	      players[playerCount].name = s; 

	      if(!(in >> players[playerCount].username)) Error("player username?");

	      playerCount++;
	    }
	}
    }

  //adjusting the starting game_seed depending on the type of game:
  switch(gameType)
    {
    case 1:
      break;
    case 2:
      game_seed += 100;
      break;
    case 3:
      game_seed += 200;
      break;
    default:
      Error("invalid game type");
    }

  //Print out all the read-in player information  
  displayPlayers(single, games_per_pair, playerCount); 
  
  //Get all available computers in the lab:  
  if(opt_genComps)
    {  
      string file = argv[argi];
      s = "ruptime | grep ^ug | grep up | cat >> " + file;
      system(s.c_str());

      ofstream out(file.c_str(), ios::app);
      if(!out.is_open()) Error("unable to open the tournament file for writing");
      out << "\n%\n\n";
      out.close();
    }   
  else
    {  
      if((c = in.peek()) == EOF)  Error("no computer list?");
    }
  
  if(skipComments(in)) Error("end of file?");
  FOREVER
    {
      if((c = in.peek()) == EOF)  Error("end of file?");
      in >> s;
      if(!s.empty())
	{ 
	  if(s[0] == '%')  break;
	  if(s[0] != '#')
	    {
	      computers[compCount];
	      computers[compCount].index = compCount;
	      computers[compCount].name = s;
	      computers[compCount].taskCount = 0; 
	      computers[compCount].isUp = true;   //should check it is up before you try to use it
	      cout << "COMP:: " << computers[compCount].name << endl << endl;
	      ++compCount;
	      getline(in,s);                       //get rid of rest of line info: how is this working for false case????
	   
	    }
	}
    }

      
  //ignore game results section for now:
  in.close();
  
  //Generate the game queue:
  gameCount = buildGameList(games_per_pair, single, playerCount, game_seed, gameType);
  displayGames(gameCount, gameType);

  threads_game = new pthread_t[gameCount];
  pthread_attr_t attr;
  pthread_mutex_init(&mutexGames, NULL);
  pthread_attr_init(&attr);
    pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  /*************MAIN LOOP*************/

  //here sleep a second & then scan the list for done games. 
  //write to disk files & grep
  //for res string & put these in the config file
  bool allDone = false;
  while(!allDone)
  {
    //checkGames(gameCount);    printf("start\n");
    allDone = checkComputers(compCount, gameCount);
    sleep(1);
    updateConfig(argv[argi], single, playerCount, gameCount, gameType);

  }		      
  
  // ***COME BACK: VALGRIND, STILL MAY LOSE MEM SOMEWHERE b/c exit()
  // doesn't call destructors

  pthread_attr_destroy(&attr);  //????
  pthread_mutex_destroy(&mutexGames);
  pthread_exit(NULL);           //??????????????

  return 0;
  
}
