/*
 * $Id: map.h,v 1.1 2006/12/31 04:09:41 orts_furtak Exp $
 */

// HOG File

#ifndef MAP_H
#define MAP_H

#include <cassert>
#include <cmath>
#include <cstdarg>
#include <cstdio>
#include <unistd.h>
#include <iostream>

#ifdef OS_MAC
#include <OpenGL/gl.h>
#else
#include <GL/gl.h>
#include <GL/glut.h>
#endif

#include "graph.h"
#include "glUtil.h"

namespace HOG{

enum tDisplay {
  kPolygons,
  kLines,
  kPoints
};

// types of ground
//  bit definition:
//  low 4 bits are subcategories
//  high bits are large categories

const int terrainBits = 4;

enum tTerrain {
  kOutOfBounds =0x0, // not part of map
  kOutOfBounds2=0x1, // not part of map - different color
  kWater=0x10,     // water
  kGround=0x20,     // ground
  kSwamp=0x21,
	kGrass=0x22,
	kBlight=0x23,
  kTrees=0x30,
  kUndefined // mixed type due to split tile
};

// ground split types
enum tSplit {
  kNoSplit=0,
  kForwardSplit=1, // ie like forward slash
  kBackwardSplit=2     // ie like back slash
};

// specifying which side of the split you want
enum tSplitSide {
  kWholeTile = 1,
  kLeftSide = 2,
  kRightSide = 3
};

enum tEdge {
  kInternalEdge,
  kLeftEdge,
  kRightEdge,
  kTopEdge,
  kBottomEdge
};

// other constants
enum {
	kUndefinedHeight = -999,
};

const int kNoGraphNode = -1;

// corner types
enum tCorner {
  kNone = 0,
  kTopLeft = 1,
  kTopRight = 2,
  kBottomLeft = 3,
  kBottomRight = 4
};

class halfTile {
public:
  halfTile();

  long corners[3];
  tTerrain type;
  long node;
};

class Tile {
public:
	Tile();
	halfTile tile1, tile2;
	tSplit split;
};

/*
 *
 * All the set/get functions should be fairly obvious, with the following
 * behavior:
 *
 * Setting the height/type without a split for a tile will reset it to be a
 * non-split tile.
 *
 * If you specify a split for a tile that is not split, 
 */

/**
 * A tile-based representation of the world.
 */

class Map {
public:
  Map(long width, long height);
  Map(char *filename);
	Map(Map *);
  Map(std::istringstream &data);
  ~Map();
  void load(char *filename);
	void setSizeMultipler(int _sizeMultiplier)
	{ sizeMultiplier = _sizeMultiplier; }
	void scale(long newWidth, long newHeight);
  void save(std::stringstream &data);
  void save(char *filename);
	Map *clone() { return new Map(this); }
	
  void print(int scale = 1);
	/** return the width of the map */
  inline long getMapWidth() const { return width; }
	/** return the height of the map */
  inline long getMapHeight() const { return height; }
	
  Tile &getTile(long x, long y);
	
  tSplit getSplit(long x, long y);
  void setSplit(long x, long y, tSplit split);


  // returns kUndefined if tile is split and you request kWholeTile
  long getTerrainType(long x, long y, tSplitSide split = kWholeTile) const;
  long getTerrainType(long x, long y, tEdge side) const;

  // if tile is not split and you specify a split side, nothing happens
  // if tile is split and you specify kWholeTile, the split remains,
  // and the terrain is applied to both sides
  void setTerrainType(long x, long y, tTerrain type, tSplitSide split = kWholeTile);

  // returns kUndefinedHeight if the tile is split and you specify
  // the whole tile
  long getHeight(long x, long y, tSplitSide split = kWholeTile);

  // if you specify a split, and the tile isn't split that direction,
  // the tile will be split that direction, and then the height applied
  void setHeight(long x, long y, long height, tSplitSide split = kWholeTile);

  // returns kUndefinedHeight if the split is inconsistant with the tile type
  long getCornerHeight(long x, long y, tCorner which, tEdge edge);
  long getCornerHeight(long x, long y, tCorner which, tSplitSide split = kWholeTile);
  void setCornerHeight(long x, long y, tCorner which, long height, tSplitSide split = kWholeTile);


  // ===============================================
  void smoothSetRectHeight(long x1, long y1, long x2, long y2, long h, tTerrain type = kGround);
  void setRectHeight(long x1, long y1, long x2, long y2, long h, tTerrain type = kGround);
  bool adjacentEdges(long x, long y, tEdge edge);
  bool adjacentCorners(long x, long y, tCorner corner);
  // returns whether we can step between two locations or not
  bool canStep(long x1, long y1, long x2, long y2);
  
  void openGLDraw(tDisplay how);
  void getOpenGLCoord(int _x, int _y, GLdouble &x, GLdouble &y, GLdouble &z, GLdouble &radius);
	void getPointFromCoordinate(point3d loc, int &px, int &py);
	double getCoordinateScale();

  void setDrawLand(bool land);
  void setDrawLandWhite(bool land) { drawLandWhite = land; }
  bool getDrawLand() { return drawLand; }
  bool getDrawLandWhite() { return drawLandWhite; }
  void drawTile(Tile *t, int x, int y, tDisplay how);
  void doVertexColor(tTerrain type, int height, bool darken = false);
  void doNormal(tSplit split, halfTile *t, int x, int y);

  float getEdgeWidthX(int x, int y);
  float getEdgeWidthY(int x, int y);

  int getNodeNum(int x, int y, tCorner c = kNone);
  void setNodeNum(int num, int x, int y, tCorner c = kNone);
  int getRevision() { return revision; }
private:
	void loadOctile(FILE *f, int height, int width);
	void loadOctileCorner(FILE *f, int height, int width);
	bool tryLoadRollingStone(FILE *f);
	bool isLegalStone(char c);
	void paintRoomInside(int x, int y);
	int width, height;
  Tile **land;
  bool drawLand;
  bool drawLandWhite;
  GLuint dList;
  bool updated;
	int sizeMultiplier;
  int revision;
};

void makeMaze(Map *map, int pathSize = 1);
}
#endif
